import heapq
from functools import reduce
from typing import Iterator, Iterable, Union

import deprecation

from .order import topological_sorting, check_topologically_sorted
from .io import isets2bas, bas2isets, to_absolute_number
from .indices import delta_stability_by_extents
from .base_functions import extension

from skmine.itemsets import LCM
from bitarray import bitarray, frozenbitarray as fbarray
from bitarray.util import zeros as bazeros, subset as basubset
from collections import deque
from tqdm.auto import tqdm


def list_intents_via_LCM(itemsets: list[fbarray], min_supp: Union[int, float] = 0, n_jobs: int = 1) -> list[fbarray]:
    """Get the list of intents by running LCM algorithm from scikit-mine

    Parameters
    ----------
    itemsets:
        The list of itemsets representing the dataset
    min_supp:
        Minimal support for the intent
    n_jobs:
        Number of jobs for LCM algorithm

    Returns
    -------
    intents:
        THe found intents

    """
    min_supp = int(min_supp * len(itemsets) if isinstance(min_supp, float) else min_supp)
    n_attrs = len(itemsets[0])

    lcm = LCM(min_supp=max(min_supp, 1), n_jobs=n_jobs)
    intents = lcm.fit_transform(list(bas2isets(itemsets)))['itemset']
    intents = list(isets2bas(intents, n_attrs))
    intents = topological_sorting(intents)[0]

    biggest_intent_support = sum(itemset.all() for itemset in itemsets)
    if biggest_intent_support >= min_supp:
        if not intents[-1].all():
            intents.append(fbarray(~bazeros(n_attrs)))

    smallest_intent = reduce(bitarray.__and__, itemsets, ~bazeros(n_attrs))
    if intents[0] != smallest_intent:
        intents.insert(0, fbarray(smallest_intent))
    return intents


def list_intents_via_Lindig(itemsets: list[bitarray], attr_extents: list[bitarray]) -> list[bitarray]:
    """Get the list of intents of itemsets grouped by equivalent classes running Lindig algorithm
    from "Fast Concept Analysis" by Christian Lindig, Harvard University, Division of Engineering and Applied Sciences

    Parameters
    ----------
    itemsets:
        list of bitarrays representing objects descriptions
    attr_extents:
        list of bitarrays representing extents of attributes

    Returns
    -------
    Lattice_data_intents:
        the list of intents of itemsets grouped by equivalent classes

    """

    class NotFound(Exception):
        pass

    def __down__(intent: list[bitarray], itemsets: list[bitarray]):
        if intent == []:
            return itemsets
        down = intent[0]
        for attr in intent[1:]:
            down = down & attr
        down =  [itemsets[i] for i in range(len(down)) if down[i] == 1]
        return down

    def __up__(extent: list[bitarray], attr_extents: list[bitarray]):
        if extent == []:
            return(attr_extents)
        up = extent[0]
        for obj in extent[1:]:
            up = up & obj
        up = [attr_extents[i] for i in range(len(up)) if up[i] == 1]
        return up

    def check_intersection(list1: list[bitarray], list2: list[bitarray]):
        has_intersection = False
        for bitarray1 in list1:
            for bitarray2 in list2:
                if bitarray1 == bitarray2:
                    has_intersection = True
                    break
        return has_intersection
    
    def compute_extent_bit(extent: list[bitarray], attr_extents: list[bitarray]):
        if extent == []:
            return(bitarray([1 for _ in range(len(attr_extents))]))
        bit_extent = extent[0]
        for obj in extent[1:]:
            bit_extent = bit_extent & obj
        return(bit_extent)
    
    def find_upper_neighbors(concept_extent: list[bitarray], itemsets: list[bitarray], attr_extents: list[bitarray]):
        min_set = [obj for obj in itemsets if obj not in concept_extent]
        neighbors = []
        for g in [obj for obj in itemsets if obj not in concept_extent]:
            B1 = __up__(concept_extent + [g], attr_extents)
            A1 = __down__(B1, itemsets)
            if not check_intersection(min_set, [obj for obj in A1 if obj not in concept_extent and obj not in [g]]):
                neighbors.append(A1)
            else:
                min_set.remove(g)
        return neighbors
    
    def find_next_concept_extent(concept_extent: list[bitarray], List_extents: list[bitarray], attr_extents: list[bitarray]):
        next_concept_extent = None
        for extent in List_extents:
            if compute_extent_bit(extent, attr_extents) < compute_extent_bit(concept_extent, attr_extents) and (next_concept_extent is None or compute_extent_bit(extent, attr_extents) > compute_extent_bit(next_concept_extent, attr_extents)):
                next_concept_extent = extent
        if next_concept_extent is not None:
            return next_concept_extent
        raise NotFound("Next concept not found in Lattice")
     
    Lattice_data_intents = []  # concepts set
    concept_extent = __down__(attr_extents, itemsets)  # Initial Concept
    Lattice_data_intents.append(concept_extent)  # Insert the initial concept into Lattice

    while True:
        for parent in find_upper_neighbors(concept_extent, itemsets, attr_extents):
            if parent not in Lattice_data_intents:
              Lattice_data_intents.append(parent)
        try:
            concept_extent = find_next_concept_extent(concept_extent, Lattice_data_intents, attr_extents)
        except NotFound:
            break
    for i in range(len(Lattice_data_intents)):
      Lattice_data_intents[i] = compute_extent_bit(Lattice_data_intents[i], attr_extents)
        
    return Lattice_data_intents


def list_attribute_concepts(intents: list[fbarray]) -> list[int]:
    """Get the indices of `intents` selected by each sole attribute"""
    assert (len(a) <= len(b) for a, b in zip(intents, intents[1:])),\
        'The list of `intents` should be topologically sorted. With the cardinality minimal intent being at the start'

    attr_concepts = [-1] * len(intents[0])
    found_attrs = bazeros(len(intents[0]))
    for intent_i, intent in enumerate(intents):
        for m in intent.itersearch(True):
            if attr_concepts[m] == -1:
                attr_concepts[m] = intent_i
        found_attrs |= intent
        if found_attrs.all():
            break

    return attr_concepts


def iter_equivalence_class(attribute_extents: list[fbarray], intent: fbarray = None) -> Iterator[fbarray]:
    """Iterate subsets of attributes from equivalence class

    The output equivalence class goes from the maximal subsets of attributes to the smallest ones.
    Equivalent subsets of attributes are the ones that describe the same subset of objects.


    Parameters
    ----------
    attribute_extents:
        The list of objects described by each specific attribute (converted to bitarrays)
    intent:
        Intent to compute equivalence class for. If None is passed, Iterate equivalence class of all attributes

    Returns
    -------
    Iterator[frozenbitarray]:
        Iterator over bitarrays representing equivalent subsets of attributes

    """
    N_OBJS, N_ATTRS = len(attribute_extents[0]), len(attribute_extents)

    intent = bazeros(N_ATTRS) if intent is None else intent

    def conjunct_extent(premise: fbarray) -> fbarray:
        res = ~bazeros(N_OBJS)
        for m in premise.itersearch(True):
            res &= attribute_extents[m]
            if not res.any():
                break

        return fbarray(res)

    total_extent = conjunct_extent(intent)
    stack = [[m] for m in intent.itersearch(True)][::-1]

    yield intent
    while stack:
        attrs_to_remove = stack.pop(0)
        last_attr = attrs_to_remove[-1]

        attrs_to_eval = bitarray(intent)
        for m in attrs_to_remove:
            attrs_to_eval[m] = False
        attrs_to_eval = fbarray(attrs_to_eval)

        conj = conjunct_extent(attrs_to_eval)
        if conj != total_extent:
            continue

        # conj == total_extent
        yield attrs_to_eval
        stack += [attrs_to_remove+[m] for m in intent.itersearch(True) if m > last_attr][::-1]


def list_keys_via_eqclass(equiv_class: Iterable[fbarray]) -> list[fbarray]:
    """List minimal subsets from given equivalence class"""
    potent_keys = []
    for new_key in equiv_class:
        potent_keys = [key for key in potent_keys if new_key & key != new_key]
        potent_keys.append(new_key)
    return potent_keys


def list_passkeys_via_eqclass(equiv_class: Iterable[fbarray]) -> list[fbarray]:
    """List subsets of minimal size from given equivalence class"""
    passkeys = []
    for descr in equiv_class:
        if not passkeys or descr.count() == passkeys[-1].count():
            passkeys.append(descr)

        if descr.count() < passkeys[-1].count():
            passkeys = [descr]
    return passkeys


def iter_keys_of_intent(intent: fbarray, attr_extents: list[fbarray]) -> Iterator[fbarray]:
    n_attrs = len(intent)
    single_attrs = list(isets2bas([{attr_idx} for attr_idx in range(n_attrs)], n_attrs))
    extent = extension(intent, attr_extents)

    def subdescriptions(description: fbarray) -> Iterator[fbarray]:
        return (description & ~single_attrs[m_i] for m_i in description.itersearch(True))

    key_candidates = deque([intent])
    while key_candidates:
        key_candidate = key_candidates.popleft()
        equiv_subdescrs = [descr for descr in subdescriptions(key_candidate)
                           if extension(descr, attr_extents) == extent]
        if not equiv_subdescrs:
            yield key_candidate
            continue

        key_candidates.extend(equiv_subdescrs)


def iter_keys_of_intent_pretentious(intent: fbarray, attr_extents: list[fbarray]) -> list[fbarray]:
    # TODO: Test the function and check if it works right and faster than the straightforward solution
    n_attrs = len(intent)
    intent = set(intent.search(True))
    extent = extension(intent, attr_extents)

    # The cycle structure is inspired by Talky-G algorithm and reverse pre-order traversal by L. Szathmary et al.
    key_candidates = [intent]
    sub_descrs_to_remove = deque([{attr_idx} for attr_idx in sorted(intent)
                                  if extension(intent-{attr_idx}, attr_extents) == extent])
    while sub_descrs_to_remove:
        attrs_to_remove = sub_descrs_to_remove.pop()
        key_candidate = intent - attrs_to_remove

        rightmost_attr = max(attrs_to_remove) if attrs_to_remove else -1
        attrs_to_remove_next = [attr_idx for attr_idx in intent
                                if rightmost_attr < attr_idx
                                and extension(key_candidate - {attr_idx}, attr_extents) == extent]
        if not attrs_to_remove_next:
            key_candidates.append(key_candidate)
            continue

        sub_descrs_to_remove.extend([attrs_to_remove | {attr_idx} for attr_idx in attrs_to_remove_next])

    single_attrs = list(isets2bas([{attr_idx} for attr_idx in range(n_attrs)], n_attrs))
    key_candidates = [reduce(fbarray.__or__, map(single_attrs.__getitem__, candidate), fbarray(bazeros(n_attrs)))
                      for candidate in key_candidates]
    key_candidates = deque(sorted(key_candidates, key=lambda candidate: candidate.count()))
    for i in range(len(key_candidates)):
        if i >= len(key_candidates):
            break

        key = key_candidates[i]
        j = i+1
        while j < len(key_candidates):
            if basubset(key, key_candidates[j]):
                del key_candidates[j]
                continue
            j += 1

    return list(key_candidates)


def list_keys(intents: list[fbarray], only_passkeys: bool = False) -> dict[fbarray, int]:
    """List all keys for all intents (i.e. minimal subsets of attributes selecting specific subsets of objects)"""
    assert all(a.count() <= b.count() for a, b in zip(intents, intents[1:])), \
        'The `intents` list should be topologically sorted by ascending order'

    n_attrs, n_intents = len(intents[-1]), len(intents)
    attrs_descendants = [bazeros(n_intents) for _ in range(n_attrs)]
    for intent_i, intent in enumerate(intents):
        for m in intent.itersearch(True):
            attrs_descendants[m][intent_i] = True

    # assuming that every subset of a key is a key => extending not-a-key cannot result in a key
    # and every subset of a passkey is a passkey

    single_attrs = list(isets2bas([[m] for m in range(n_attrs)], n_attrs))

    if only_passkeys:
        passkey_sizes = [n_attrs] * n_intents

    keys_dict = {fbarray(bazeros(n_attrs)): 0}
    attrs_to_test = deque([m_ba for m_ba in single_attrs])
    while attrs_to_test:
        attrs = attrs_to_test.popleft()
        attrs_indices = list(attrs.itersearch(True))

        if any(attrs & (~single_attrs[m]) not in keys_dict for m in attrs_indices):
            continue

        common_descendants = ~bazeros(n_intents)
        for m in attrs.itersearch(True):
            common_descendants &= attrs_descendants[m]
        if not common_descendants.any():
            continue
        meet_intent_idx = common_descendants.index(True)

        if only_passkeys:
            # `attrs` is not a passkey because of the size
            if passkey_sizes[meet_intent_idx] < attrs.count():
                continue

        # if subset of attrs is not a key, or a key of the same intent
        if any(keys_dict[attrs & (~single_attrs[m])] == meet_intent_idx for m in attrs_indices):
            continue

        keys_dict[attrs] = meet_intent_idx
        if only_passkeys:
            passkey_sizes[meet_intent_idx] = attrs.count()
        if not intents[meet_intent_idx].all():
            max_attr_idx = attrs_indices[-1] if attrs_indices else -1
            attrs_to_test.extend([attrs | m_ba for m_ba in single_attrs[max_attr_idx+1:]])

    return keys_dict


def list_passkeys(intents: list[fbarray]) -> dict[fbarray, int]:
    """List all passkeys for all intents

     (i.e. subsets of attributes of minimal size selecting specific subsets of objects)"""
    return list_keys(intents, only_passkeys=True)


def list_keys_for_extents(
        extents: list[fbarray], attr_extents: list[fbarray],
        only_passkeys: bool = False
) -> dict[fbarray, int]:
    """List all keys for given extents (i.e. minimal subsets of attributes selecting given closed subsets of objects)"""
    if not check_topologically_sorted(extents, ascending=False):
        extents, orig_to_topsort_idx_map = topological_sorting(extents, ascending=False)
    else:
        orig_to_topsort_idx_map = list(range(len(extents)))
    topsort_to_orig_idx_map = {sort_idx: orig_idx for orig_idx, sort_idx in enumerate(orig_to_topsort_idx_map)}

    extents_idx_map: dict[fbarray, int] = {extent: extent_i for extent_i, extent in enumerate(extents)}
    min_support = extents[-1].count()

    n_attrs, n_objs, n_extents = len(attr_extents), len(attr_extents[0]), len(extents)
    total_extent = attr_extents[0] | ~attr_extents[0]

    single_attrs = list(isets2bas([[m] for m in range(n_attrs)], n_attrs))

    if only_passkeys:
        passkey_sizes = [n_attrs] * n_extents

    keys_dict: dict[fbarray, tuple[int, int|None]] = {fbarray(bazeros(n_attrs)): (n_objs, extents_idx_map.get(total_extent))}
    testing_stack = deque(single_attrs)  # iteration order is inspired by Talky-G algorithm
    while testing_stack:
        attrs = testing_stack.pop()
        attrs_indices = list(attrs.itersearch(True))
        sub_descriptions = [attrs & (~single_attrs[m]) for m in attrs_indices]

        # check that every subset of attrs is a key
        if any(subattrs not in keys_dict for subattrs in sub_descriptions):
            continue

        curr_extent: fbarray = reduce(fbarray.__and__, map(attr_extents.__getitem__, attrs_indices), total_extent)
        curr_extent_i = extents_idx_map[curr_extent] if curr_extent in extents_idx_map else None

        if only_passkeys and curr_extent_i is not None:
            # `attrs` is not a passkey because of the size
            if passkey_sizes[curr_extent_i] < attrs.count():
                continue

        # if subset of attrs has the same support then it is not a key
        support = curr_extent.count()
        if any(keys_dict[subattrs][0] <= support for subattrs in sub_descriptions):
            continue

        keys_dict[attrs] = (support, curr_extent_i)
        if only_passkeys and curr_extent_i is not None:
            passkey_sizes[curr_extent_i] = attrs.count()

        has_smaller_extents = support > min_support
        if curr_extent_i is not None and has_smaller_extents:
            has_smaller_extents = any(basubset(ext, curr_extent) for ext in extents[curr_extent_i+1:])

        if has_smaller_extents:
            max_attr_idx = attrs_indices[-1] if attrs_indices else -1
            testing_stack.extend([attrs | m_ba for m_ba in single_attrs[max_attr_idx+1:]])

    return {key: topsort_to_orig_idx_map[extent_i]
            for key, (support, extent_i) in keys_dict.items() if extent_i is not None}


def list_passkeys_for_extents(
        extents: list[fbarray], attr_extents: list[fbarray],
) -> dict[fbarray, int]:
    return list_keys_for_extents(extents, attr_extents, only_passkeys=True)


@deprecation.deprecated(deprecated_in="0.2.0", removed_in="0.2.1",
                        details="Use `list_stable_extents_via_gsofia` function instead. It is faster and more reliable")
def list_stable_extents_via_sofia(
        attribute_extents: Iterable[fbarray],
        n_stable_extents: int, min_supp: int = 0,
        use_tqdm: bool = False, n_attributes: int = None
) -> set[fbarray]:
    stable_extents = set()
    for attr_extent in tqdm(attribute_extents, disable=not use_tqdm, total=n_attributes):
        if not stable_extents:  # Set up the top extent, as soon as we know the number of objects
            stable_extents.add(fbarray(~bazeros(len(attr_extent))))

        if attr_extent.count() < min_supp:
            continue

        new_extents = (extent & attr_extent for extent in stable_extents)
        new_extents = filter(lambda extent: extent.count() >= min_supp and extent not in stable_extents, new_extents)
        stable_extents |= set(new_extents)

        if len(stable_extents) > n_stable_extents:
            extents_top_sort = sorted(stable_extents, key=lambda extent: extent.count())
            delta_stabilities = list(delta_stability_by_extents(extents_top_sort))
            stab_thold = sorted(delta_stabilities)[-n_stable_extents]
            stable_extents = {extent for extent, stab in zip(extents_top_sort, delta_stabilities) if stab >= stab_thold}

    if len(stable_extents) > n_stable_extents:
        stable_extents = {extent for extent, stab in zip(extents_top_sort, delta_stabilities) if stab > stab_thold}

    return stable_extents


def list_stable_extents_via_gsofia(
        attribute_extents: Iterable[fbarray], n_objects: int = None,
        min_delta_stability: Union[int, float] = 0,
        n_stable_extents: int = None,
        min_supp: Union[int, float] = 0,
        use_tqdm: bool = False, n_attributes: int = None,
) -> set[fbarray]:
    """Select the most stable extents restricted by min. delta stability value, n. most stable extents, and min. support

    An extent is ∆-stable if there exists no sub-extent that covers ∆ fewer objects than the given extent.

    Extent is a subset of objects, described by some attributes.
    Every extent is an intersection of a subset of `attribute_extents`.


    Parameters
    ----------
    attribute_extents:
        Iterable over extents of attributes.
        Every extent is a set of objects, described by an attribute represented with a bitarray.
    n_objects:
        The number of objects in the data.
        Have to be specified unless it can be computed as `len(attribute_extents[0])`.
    min_delta_stability:
        A lower bound for delta stability of the returned extents.
        The parameter should represent either an absolute number or a percentage of objects in the data.
        Delta stability shows how many objects a description will lose if made a bit more precise.
        Delta-stable extents are the ones that are selected by delta-stable descriptions.
    n_stable_extents:
        An upper bound on the number of the returned extents.
        The bigger is the number, the longer it takes to run the algorithm and the more precise are the results.
    min_supp:
        A lower bound on the support of the returned extents.
        Support of the extent shows how many objects are contained in this extent.
    use_tqdm:
        A flag, whether to visualise the progress bar via tqdm package.
        If set to True and `attribute_extents` is a generator,
         please specify `n_attributes` parameter to define the maximal amount of steps of the algorithm.
    n_attributes:
        The number of elements in `attribute_extents` iterator.
        Optional parameter that one might need to specify in order to set up the width of tqdm progress bar
        (if `use_tqdm` is set to True, and the length of `attribute_extents` cannot be computed automatically).


    Returns
    -------
    set[frozenbitarray]:
        A set of stable extents computed w.r.t. the other parameters.

    Notes
    -----
    The algorithm falls in line with gSofia algorithm presented in
    paper "Efficient Mining of Subsample-Stable Graph Patterns"
    by A. Buzmakov, S.O. Kuznetsov, and A. Napoli from International Conference on Data Mining of 2017.
    However, we remove the graph-specific part of the algorithm to make it applicable for any type of attributes.

    """
    if n_objects is None:
        try:
            n_objects = len(attribute_extents[0])
        except TypeError:
            raise TypeError('The function cannot establish the number of objects. '
                            'Please, either provide the number of objects with `n_objects` parameter '
                            'or convert `attribute_extents` parameter into a list')
    if n_attributes is None:
        try:
            n_attributes = len(attribute_extents)
        except:
            pass
    min_delta_stability = to_absolute_number(min_delta_stability, n_objects)
    min_supp = to_absolute_number(min_supp, n_objects)

    # Create dict with a format:  extent => (delta_index, children_extents)
    top_extent = fbarray(~bazeros(n_objects))
    stable_extents: dict[fbarray, tuple[int, set[fbarray]]] = {top_extent: (n_objects, set())}

    # noinspection PyTypeChecker
    attr_extent_iterator: Iterable[fbarray] = tqdm(attribute_extents, total=n_attributes, disable=not use_tqdm)
    for attr_extent in attr_extent_iterator:
        # The following code mimics "ExtendProjection" function from gSofia algorithm
        old_stable_extents, stable_extents = dict(stable_extents), dict()
        for extent, (delta, children) in old_stable_extents.items():
            # Create new extent
            extent_new = extent & attr_extent
            if extent_new == extent:
                stable_extents[extent] = (delta, children)
                continue

            # Update the stability of the old extent given its new child: `extent_new`
            delta = min(delta, extent.count() - extent_new.count())
            if delta >= min_delta_stability:
                stable_extents[extent] = (delta, children | {extent_new})

            # Skip the new extent if it is too small
            if extent_new.count() < min_supp:
                continue

            # Find the delta-index of the new extent and its children extents
            delta_new, children_new = extent_new.count(), []
            for child in children:
                child_new = child & attr_extent
                delta_new = min(delta_new, extent_new.count() - child_new.count())
                if delta_new < min_delta_stability:
                    children_new = []
                    break
                children_new.append(child_new)

            # Skip the new extent if it is too unstable
            if delta_new < min_delta_stability:
                continue

            # Filter the maximal children (so filter out transitive children that are children to other children)
            children_new, i = sorted(children_new, key=lambda ext: ext.count(), reverse=True), 0
            while i < len(children_new):
                child = children_new[i]
                has_bigger_child = any(basubset(child, bigger_child) for bigger_child in children_new[:i])
                if has_bigger_child:
                    del children_new[i]
                else:
                    i += 1
            children_new = set(children_new)

            # At this point, we know that `extent_new` is big enough and is stable enough
            stable_extents[extent_new] = (delta_new, children_new)

        # If only the best stable concepts required
        if n_stable_extents is not None and len(stable_extents) > n_stable_extents:
            most_stable_extents = heapq.nlargest(n_stable_extents, stable_extents.items(), key=lambda x: x[1][0])
            thold = most_stable_extents[-1][1][0]
            n_border_stable = 0
            for (_, (stab, _)) in most_stable_extents[::-1]:
                if stab > thold:
                    continue
                n_border_stable += 1
            n_border_stable_total = sum(stab == thold for _, (stab, _) in stable_extents.items())
            if n_border_stable < n_border_stable_total:
                most_stable_extents = most_stable_extents[:-n_border_stable]
            stable_extents = dict(most_stable_extents)

    return set(stable_extents)



from caspailleur import base_functions as bfunc


def test_powerset():
    iterable = (1, 2, 3)
    pset_true = [[], [1], [2], [3], [1, 2], [1, 3], [2, 3], [1, 2, 3]]
    pset_true = [frozenset(ar) for ar in pset_true]

    pset = bfunc.powerset(iterable)
    assert list(pset) == pset_true


def test_is_subset_of():
    assert bfunc.is_subset_of(set(), {1, 2, 3})
    assert bfunc.is_subset_of({2, 5, 3}, {2, 1, 3, 5})
    assert not bfunc.is_subset_of({2, 1, 3, 5}, {1})
    assert bfunc.is_subset_of({1, 2, 3}, {1, 2, 3})
    assert not bfunc.is_subset_of({0, 1, 3}, {0, 1, 2})


def test_is_psubset_of():
    assert bfunc.is_psubset_of(set(), {1, 2, 3})
    assert bfunc.is_psubset_of({2, 5, 3}, {2, 1, 3, 5})
    assert not bfunc.is_psubset_of({2, 1, 3, 5}, {1})
    assert not bfunc.is_psubset_of({1, 2, 3}, {1, 2, 3})


def test_closure():
    crosses_per_columns = [
        {0, 1,    3},  # col 0
        {0, 1, 2   },  # col 1
        {0,    2   },  # col 2
        {0, 1, 2, 3},  # col 3
        set(),         # col 4
    ]

    assert list(bfunc.closure([0], crosses_per_columns)) == [0, 3]
    assert list(bfunc.closure([0, 1], crosses_per_columns)) == [0, 1, 3]
    assert list(bfunc.closure([1, 2], crosses_per_columns)) == [1, 2, 3]
    assert list(bfunc.closure([], crosses_per_columns)) == [3]



import pytest
import numpy as np

from caspailleur import mine_equivalence_classes as mec
from caspailleur import io

from bitarray import frozenbitarray as fbarray
from bitarray.util import zeros as bazeros
from bitarray import bitarray


def test_list_intents_via_LCM():
    itemsets = [
        {0, 3},
        {0, 2},
        {1, 2},
        {1, 2, 3}
    ]

    intents_true = list(io.isets2bas([set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}], 5))

    intents = mec.list_intents_via_LCM(list(io.isets2bas(itemsets, 5)))
    assert intents == intents_true

    freq_intents = mec.list_intents_via_LCM(list(io.isets2bas(itemsets, 5)), min_supp=2)
    freq_intents_true = list(io.isets2bas([set(), {0}, {2}, {3}, {1, 2}], 5))
    assert freq_intents == freq_intents_true
    assert mec.list_intents_via_LCM(list(io.isets2bas(itemsets, 5)), min_supp=0.5) == freq_intents_true

    K = np.array([[True, False, False, False, False, True], [False, True, False, False, False, False]])
    itemsets = io.np2bas(K)
    intents_true = list(io.isets2bas([set(), {1}, {0, 5}, {0, 1, 2, 3, 4, 5}], 6))

    intents = mec.list_intents_via_LCM(itemsets)
    assert intents == intents_true


def test_list_intents_via_lindig():
    K = np.array([
    [True, False, False, True],
    [True, False, True, False],
    [False, True, True, False],
    [False, True, True, True]])

    itemsets = io.np2bas(K)
    attr_extents = io.np2bas(K.T)
    intents_true = ['1111', '1001', '1010', '0111', '1000', '0010', '0001', '0000', '0110']
 
    intents_true = [bitarray(x) for x in intents_true]

    intents = mec.list_intents_via_Lindig(itemsets, attr_extents)
    assert intents == intents_true


def test_list_attribute_concepts():
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3},  {0, 1, 2, 3, 4}]
    attr_concepts_true = [1, 6, 2, 3, 8]

    attr_concepts = mec.list_attribute_concepts(list(io.isets2bas(intents, 5)))
    assert attr_concepts == attr_concepts_true


def tests_iter_equivalence_class():
    K = np.array([
        [True, False, False, True, False],
        [True, False, True, False, False],
        [False, True, True, False, False],
        [False, True, True, True, False],
    ])
    attr_extents = list(io.np2bas(K.T))
    eq_class_true = [
        {0, 1, 2, 3, 4}, {0, 1, 2, 3}, {0, 1, 2, 4}, {0, 1, 3, 4}, {0, 2, 3, 4}, {1, 2, 3, 4},
        {0, 1, 2}, {0, 1, 3}, {0, 1, 4}, {0, 2, 3}, {0, 2, 4}, {0, 3, 4}, {1, 2, 4}, {1, 3, 4}, {2, 3, 4},
        {0, 1}, {0, 4}, {1, 4}, {2, 4}, {3, 4}, {4}
    ]

    eq_class = list(mec.iter_equivalence_class(attr_extents, ~fbarray(bazeros(5))))
    assert eq_class == list(io.isets2bas(eq_class_true, 5))


def test_list_keys_via_eqclass():
    eq_class = list(io.isets2bas([
        {0, 1, 2, 3, 4}, {0, 1, 2, 3}, {0, 1, 2, 4}, {0, 1, 3, 4}, {0, 2, 3, 4}, {1, 2, 3, 4},
        {0, 1, 2}, {0, 1, 3}, {0, 1, 4}, {0, 2, 3}, {0, 2, 4}, {0, 3, 4}, {1, 2, 4}, {1, 3, 4}, {2, 3, 4},
        {0, 1}, {0, 4}, {1, 4}, {2, 4}, {3, 4}, {4}
    ], 5))
    keys_true = list(io.isets2bas([{0, 2, 3}, {0, 1}, {4}], 5))

    keys = mec.list_keys_via_eqclass(eq_class)
    assert keys == keys_true


def test_list_passkeys_via_keys():
    keys = [{0, 2, 3}, {0, 1}, {4}, {0, 2, 4}]
    pkeys_true = [{4}]

    pkeys = mec.list_passkeys_via_eqclass(io.isets2bas(keys, 5))
    assert pkeys == list(io.isets2bas(pkeys_true, 5))


def test_iter_keys_of_intent():
    n_attrs = 5
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    intents = list(io.isets2bas(intents, n_attrs))

    attr_extents = [{0, 1}, {2, 3}, {0, 2, 3}, {1, 2}, set()]
    attr_extents = list(io.isets2bas(attr_extents, n_attrs))

    keys_true = [
        [set()],  # intent: set()
        [{0}],  # intent: {0}
        [{2}],  # intent: {2}
        [{3}],  # intent: {3}
        [{0, 2}],  # intent: {0, 2}
        [{0, 3}],  # intent: {0, 3}
        [{1}],  # intent: {1, 2}
        [{1, 3}, {2, 3}],  # intent: {1, 2, 3}
        [{4}, {0, 1}, {0, 2, 3}],  # intent: {0, 1, 2, 3, 4}
    ]
    keys_true = [list(io.isets2bas(keys, n_attrs)) for keys in keys_true]

    for key_list_true, intent in zip(keys_true, intents):
        key_list = list(mec.iter_keys_of_intent(intent, attr_extents))
        assert set(key_list) == set(key_list_true), f'Problematic intent: {intent}'


def test_iter_keys_of_intent_pretentious():
    n_attrs = 5
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    intents = list(io.isets2bas(intents, n_attrs))

    attr_extents = [{0, 1}, {2, 3}, {0, 2, 3}, {1, 2}, set()]
    attr_extents = list(io.isets2bas(attr_extents, n_attrs))

    keys_true = [
        [set()],  # intent: set()
        [{0}],  # intent: {0}
        [{2}],  # intent: {2}
        [{3}],  # intent: {3}
        [{0, 2}],  # intent: {0, 2}
        [{0, 3}],  # intent: {0, 3}
        [{1}],  # intent: {1, 2}
        [{1, 3}, {2, 3}],  # intent: {1, 2, 3}
        [{4}, {0, 1}, {0, 2, 3}],  # intent: {0, 1, 2, 3, 4}
    ]
    keys_true = [list(io.isets2bas(keys, n_attrs)) for keys in keys_true]

    for key_list_true, intent in zip(keys_true, intents):
        key_list = list(mec.iter_keys_of_intent_pretentious(intent, attr_extents))
        assert set(key_list) == set(key_list_true), f'Problematic intent: {intent}'


def test_list_keys():
    n_attrs = 5
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    intents = list(io.isets2bas(intents, n_attrs))

    keys_true = [
        (set(), 0),
        ({0}, 1), ({1}, 6), ({2}, 2), ({3}, 3), ({4}, 8),
        ({0, 1}, 8), ({0, 2}, 4), ({0, 3}, 5), ({1, 3}, 7), ({2, 3}, 7),
        ({0, 2, 3}, 8)
    ]
    keys_true = {next(io.isets2bas([key], n_attrs)): intent_i for key, intent_i in keys_true}

    keys = mec.list_keys(intents)
    assert keys == keys_true

    with pytest.raises(AssertionError):
        mec.list_keys(intents[::-1])


def test_list_passkeys():
    n_attrs = 5
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    intents = list(io.isets2bas(intents, n_attrs))

    pkeys_true = [
        (set(), 0),
        ({0}, 1), ({1}, 6), ({2}, 2), ({3}, 3), ({4}, 8),
        ({0, 2}, 4), ({0, 3}, 5), ({1, 3}, 7), ({2, 3}, 7),
    ]
    pkeys_true = {next(io.isets2bas([key], n_attrs)): intent_i for key, intent_i in pkeys_true}

    pkeys = mec.list_passkeys(intents)
    assert pkeys == pkeys_true

    with pytest.raises(AssertionError):
        mec.list_passkeys(intents[::-1])


def test_list_keys_for_extents():
    attr_extents = [{0, 1}, {2, 3}, {0, 2, 3}, {1, 2}, set()]
    extents = [{0, 1, 2, 3},
               {0, 1}, {0, 2, 3}, {1, 2},
               {0}, {1}, {2, 3}, {2}, set()]
    keys_true = [
        (set(), 0),
        ({0}, 1), ({1}, 6), ({2}, 2), ({3}, 3), ({4}, 8),
        ({0, 1}, 8), ({0, 2}, 4), ({0, 3}, 5), ({1, 3}, 7), ({2, 3}, 7),
        ({0, 2, 3}, 8)
    ]

    n_attrs, n_objs = len(attr_extents), len(extents[0])
    attr_extents, extents = [list(io.isets2bas(isets, n_objs)) for isets in [attr_extents, extents]]
    keys_true = {next(io.isets2bas([key], n_attrs)): intent_i for key, intent_i in keys_true}

    keys = mec.list_keys_for_extents(extents, attr_extents)
    assert keys == keys_true

    # Partially Ordered Set case
    attr_extents = [{0, 1, 2}, {0, 1, 3}, {0}, {0}, {1}]
    extents = [{0, 1, 2}, {0, 1, 3}, {0}, {1}, set()]
    keys_true = [
        ({0}, 0), ({1}, 1), ({2}, 2), ({3}, 2),
        ({4}, 3), ({2, 4}, 4), ({3, 4}, 4)
    ]
    n_attrs, n_objs = len(attr_extents), 4
    attr_extents, extents = [list(io.isets2bas(isets, n_objs)) for isets in [attr_extents, extents]]
    keys_true = {next(io.isets2bas([key], n_attrs)): intent_i for key, intent_i in keys_true}

    keys = mec.list_keys_for_extents(extents, attr_extents)
    assert keys == keys_true


def test_list_stable_extents_via_sofia():
    # Data inspired by Animal Movement Context
    all_extents = [
        fbarray('1111111111111111'),  # delta_stab: 8
        fbarray('0000000111111110'),  # delta_stab: 3
        fbarray('0000111101111000'),  # delta_stab: 3
        fbarray('1011111000000000'),  # delta_stab: 3
        fbarray('0000000101111000'),  # delta_stab: 5
        fbarray('0000111000000000'),  # delta_stab: 3
        fbarray('0011000000000000'),  # delta_stab: 2
        fbarray('0000000000000000'),  # delta_stab: 0
    ]

    bin_attributes = [
        fbarray('1011111000000000'),
        fbarray('0000111101111000'),
        fbarray('0000000111111110'),
        fbarray('0011000000000000'),
    ]

    stable_extents = mec.list_stable_extents_via_sofia(bin_attributes, 8)
    assert set(all_extents) == stable_extents

    stable_extents = mec.list_stable_extents_via_sofia(bin_attributes, 8, min_supp=3)
    assert set(all_extents[:-2]) == stable_extents

    stable_extents = mec.list_stable_extents_via_sofia(bin_attributes, 6)
    assert set(all_extents[:-2]) == stable_extents

    stable_extents = mec.list_stable_extents_via_sofia(bin_attributes, 5)
    assert {all_extents[0], all_extents[4]} == stable_extents


def test_list_stable_extents_via_gsofia():
    attr_extents = [fbarray('110'),  fbarray('101'), fbarray('011')]
    extents_true = {
        fbarray('111'),
        fbarray('110'), fbarray('101'), fbarray('011'),
        fbarray('100'), fbarray('010'), fbarray('001'),
        fbarray('000')
    }

    stable_extents = mec.list_stable_extents_via_gsofia(attr_extents, min_delta_stability=0)
    assert stable_extents == extents_true

    stable_extents = mec.list_stable_extents_via_gsofia(attr_extents, min_delta_stability=1)
    assert stable_extents == extents_true - {fbarray('000')}

    stable_extents = mec.list_stable_extents_via_gsofia(attr_extents, min_delta_stability=2)
    assert stable_extents == set()

    # Data inspired by Animal Movement Context
    all_extents = [
        fbarray('1111111111111111'),  # delta_stab: 8
        fbarray('0000000111111110'),  # delta_stab: 3
        fbarray('0000111101111000'),  # delta_stab: 3
        fbarray('1011111000000000'),  # delta_stab: 3
        fbarray('0000000101111000'),  # delta_stab: 5
        fbarray('0000111000000000'),  # delta_stab: 3
        fbarray('0011000000000000'),  # delta_stab: 2
        fbarray('0000000000000000'),  # delta_stab: 0
    ]
    bin_attributes = [
        fbarray('1011111000000000'),
        fbarray('0000111101111000'),
        fbarray('0000000111111110'),
        fbarray('0011000000000000'),
    ]
    stable_extents = mec.list_stable_extents_via_gsofia(bin_attributes, min_delta_stability=0)
    assert set(all_extents) == stable_extents

    stable_extents = mec.list_stable_extents_via_gsofia(bin_attributes, min_delta_stability=3)
    assert set(all_extents[:-2]) == stable_extents

    stable_extents = mec.list_stable_extents_via_gsofia(bin_attributes, min_delta_stability=4)
    assert {all_extents[0], all_extents[4]} == stable_extents

    stable_extents = mec.list_stable_extents_via_gsofia(bin_attributes, min_delta_stability=0.25)
    assert {all_extents[0], all_extents[4]} == stable_extents

    stable_extents = mec.list_stable_extents_via_gsofia(bin_attributes, min_delta_stability=3, min_supp=6)
    assert {all_extents[0], all_extents[1], all_extents[2], all_extents[3]} == stable_extents

    stable_extents = mec.list_stable_extents_via_gsofia(bin_attributes,
                                                        min_delta_stability=3, min_supp=5, n_stable_extents=2)
    assert {all_extents[0], all_extents[4]} == stable_extents

    stable_extents = mec.list_stable_extents_via_gsofia(bin_attributes, n_stable_extents=3)
    assert {all_extents[0], all_extents[4]} == stable_extents



from .implication_bases import iter_proper_premises_via_keys, list_pseudo_intents_via_keys
from .mine_equivalence_classes import list_intents_via_LCM, list_keys, list_passkeys, list_stable_extents_via_gsofia
from .order import sort_intents_inclusion, inverse_order
from .indices import linearity_index, distributivity_index

from .api import iter_descriptions, mine_descriptions, mine_concepts, mine_implications

__version__ = '0.2.0'



from typing import Dict, Any, Union

import deprecation
import numpy as np

from . import io
from . import implication_bases as ibases
from . import mine_equivalence_classes as mec
from . import order as ordermod
from . import indices as indicesmod


@deprecation.deprecated(
    deprecated_in="0.2.0", removed_in="0.2.1",
    details="Use functions `mine_implications`, `mine_concepts` and `mine_descriptions` from API module. "
            "They provide easier to work-with output"
)
def explore_data(K: np.ndarray, min_sup: Union[int, float] = 0, return_itemsets: bool = True) -> Dict[str, Any]:
    """One function to output all dependencies in the data

    Parameters
    ----------
    K:
        Binary dataset represented with numpy bool 2D array
    min_sup:
        Minimal support for intents
    return_itemsets:
        A flag whether to return the output as sets of attribute indices or as bitarrays/
        The second option is memory efficient but less interpretable.

    Returns
    -------
    Dictionary with
        intents, keys, passkeys,
        pseudo_intents, proper_premises,
        intents_ordering,
        and linearity, distributivity indices
    """
    itemsets = list(io.np2bas(K))

    intents = mec.list_intents_via_LCM(itemsets, min_supp=min_sup)
    keys = mec.list_keys(intents)
    passkeys = mec.list_passkeys(intents)

    children_ordering = ordermod.sort_intents_inclusion(intents)
    parents_ordering = ordermod.inverse_order(children_ordering)

    pseudo_intents = dict(ibases.list_pseudo_intents_via_keys(keys.items(), intents))
    proper_premises = dict(ibases.iter_proper_premises_via_keys(intents, keys))

    n_transitive_parents = sum(tparents.count() for tparents in ordermod.close_transitive_subsumption(parents_ordering))
    linearity = indicesmod.linearity_index(n_transitive_parents, len(intents))
    distributivity = indicesmod.distributivity_index(intents, parents_ordering, n_transitive_parents)

    output = dict(
        intents=intents,
        keys=keys, passkeys=passkeys,
        pseudo_intents=pseudo_intents, proper_premises=proper_premises,
        intents_ordering=parents_ordering,
        linearity=linearity, distributivity=distributivity
    )
    if return_itemsets:
        for output_k, output_v in output.items():
            itemset_v = output_v
            if isinstance(output_v, dict):
                itemset_v = dict(zip(io.bas2isets(output_v.keys()), output_v.values()))
            if isinstance(output_v, list):
                itemset_v = list(io.bas2isets(output_v))
            output[output_k] = itemset_v

    return output



import pandas as pd

from caspailleur import api


def assert_df_equality(df1: pd.DataFrame, df2: pd.DataFrame):
    assert list(df1.index) == list(df2.index)
    assert list(df1.columns) == list(df2.columns)
    for f in df1:
        assert list(df1[f]) == list(df2[f]), f"Problematic column: {f}"


def test_mine_descriptions():
    data = {'g1': ['a', 'b'], 'g2': ['b', 'c']}

    descriptions_data_true = pd.DataFrame({
        'description': [set(), {'a'}, {'b'}, {'c'}, {'a', 'b'}, {'a', 'c'}, {'b', 'c'}, {'a', 'b', 'c'}],
        'extent': [{'g1', 'g2'}, {'g1'}, {'g1', 'g2'}, {'g2'}, {'g1'}, set(), {'g2'}, set()],
        'intent': [{'b'}, {'a', 'b'}, {'b'}, {'b', 'c'}, {'a', 'b'}, {'a', 'b', 'c'}, {'b', 'c'}, {'a', 'b', 'c'}],
        'support': [2, 1, 2, 1, 1, 0, 1, 0],
        'delta_stability': [0, 0, 1, 0, 1, 0, 1, 0],
        'is_closed': [False, False, True, False, True, False, True, True],
        'is_key': [True, True, False, True, False, True, False, False],
        'is_passkey': [True, True, False, True, False, True, False, False],
        'is_proper_premise': [True, False, False, False, False, False, False, False],
        'is_pseudo_intent': [True, False, False, False, False, False, False, False]
    })
    descriptions_data = api.mine_descriptions(data, to_compute='all')
    assert_df_equality(descriptions_data, descriptions_data_true)

    # test min_support threshold
    for min_supp in [1, 2]:
        freq_df_true = descriptions_data_true[descriptions_data_true['support'] >= min_supp].reset_index(drop=True)
        freq_df = api.mine_descriptions(data, min_support=min_supp, to_compute='all')
        assert_df_equality(freq_df, freq_df_true)

    for min_supp in [0.5, 1.0]:
        freq_df_true = descriptions_data_true[descriptions_data_true['support']/len(data) >= min_supp].reset_index(drop=True)
        freq_df = api.mine_descriptions(data, min_support=min_supp, to_compute='all')
        assert_df_equality(freq_df, freq_df_true)


def test_iter_descriptions():
    data = {'g1': ['a', 'b'], 'g2': ['b', 'c']}

    descriptions_data = list(api.iter_descriptions(data))
    assert_df_equality(pd.DataFrame(descriptions_data), api.mine_descriptions(data))

    to_compute = ['description', 'extent', 'intent', 'is_proper_premise', 'is_pseudo_intent', 'delta_stability']
    descriptions_data = list(api.iter_descriptions(data, to_compute=to_compute))
    assert_df_equality(pd.DataFrame(descriptions_data), api.mine_descriptions(data, to_compute='all')[to_compute])


def test_mine_concepts():
    data = {'g1': ['a', 'b'], 'g2': ['b', 'c']}

    concepts_df_true = pd.DataFrame({
        'extent': [{'g1', 'g2'}, {'g1'}, {'g2'}, set()],
        'intent': [{'b'}, {'a', 'b'}, {'b', 'c'}, {'a', 'b', 'c'}],
        'support': [2, 1, 1, 0],
        'delta_stability': [1, 1, 1, 0],
        'keys': [[set()], [{'a'}], [{'c'}], [{'a','c'}]],
        'passkeys': [[set()], [{'a'}], [{'c'}], [{'a', 'c'}]],
        'proper_premises': [[set()], [], [], []],
        'pseudo_intents': [[set()], [], [], []],
        'previous_concepts': [{1, 2}, {3}, {3}, set()],
        'next_concepts': [set(), {0}, {0}, {1, 2}],
        'sub_concepts': [{1, 2, 3}, {3}, {3}, set()],
        'super_concepts': [set(), {0}, {0}, {0, 1, 2}],
    })

    concepts_df = api.mine_concepts(data, to_compute='all')
    assert_df_equality(concepts_df, concepts_df_true)

    stable_concepts_df_true = concepts_df_true[:3]
    for f in ['previous_concepts', 'next_concepts', 'sub_concepts', 'super_concepts']:
        stable_concepts_df_true[f] = stable_concepts_df_true[f] - {3}

    stable_concepts_df = api.mine_concepts(data, to_compute='all', min_delta_stability=1)
    assert_df_equality(stable_concepts_df, stable_concepts_df_true)

    stable_concepts_df = api.mine_concepts(data, to_compute='all', n_stable_concepts=3)
    assert_df_equality(stable_concepts_df, stable_concepts_df_true)

    stable_concepts_df = api.mine_concepts(data, to_compute='all', min_delta_stability=1, n_stable_concepts=3)
    assert_df_equality(stable_concepts_df, stable_concepts_df_true)


def test_mine_implications():
    data = {'g1': ['a', 'b'], 'g2': ['b', 'c']}

    impls_df_true = pd.DataFrame({
        'premise': [set()],
        'conclusion': [{'b'}],
        'conclusion_full': [{'b'}],
        'extent': [{'g1', 'g2'}],
        'support': 2
    })

    impls_df = api.mine_implications(data, 'Proper Premise')
    assert_df_equality(impls_df, impls_df_true)

    impls_df = api.mine_implications(data, 'Pseudo-Intent')
    assert_df_equality(impls_df, impls_df_true)

    for basis_name in ['Duquenne-Guigues', 'Minimum', 'Canonical']:
        impls_df = api.mine_implications(data, basis_name)
        assert_df_equality(impls_df, impls_df_true)

    for basis_name in ['Canonical Direct', 'Karell']:
        impls_df = api.mine_implications(data, basis_name)
        assert_df_equality(impls_df, impls_df_true)

    impls_df_true_unit = pd.DataFrame({
        'premise': [set()],
        'conclusion': ['b'],
        'conclusion_full': [{'b'}],
        'extent': [{'g1', 'g2'}],
        'support': 2
    })
    impls_df = api.mine_implications(data, 'Proper Premise', unit_base=True)
    assert_df_equality(impls_df, impls_df_true_unit)

    impls_df = api.mine_implications(data, 'Proper Premise', unit_base=True,
                                     to_compute=['premise', 'conclusion', 'extent'])
    assert_df_equality(impls_df, impls_df_true_unit[['premise', 'conclusion', 'extent']])

    # TODO: Add refined tests for implication (and especially unit) base



import numpy as np

from caspailleur.orchestrator import explore_data
from caspailleur.indices import linearity_index, distributivity_index
from caspailleur.io import isets2bas


def test_explore_data():
    K = np.array([
        [True, False, False, True, False],
        [True, False, True, False, False],
        [False, True, True, False, False],
        [False, True, True, True, False],
    ])

    intents_true = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    intents_true = [frozenset(intent) for intent in intents_true]

    keys_true = [
        (set(), 0),
        ({0}, 1), ({1}, 6), ({2}, 2), ({3}, 3), ({4}, 8),
        ({0, 1}, 8), ({0, 2}, 4), ({0, 3}, 5), ({1, 3}, 7), ({2, 3}, 7),
        ({0, 2, 3}, 8)
    ]
    keys_true = {frozenset(key): intent_i for key, intent_i in keys_true}

    passkeys_true = [
        (set(), 0),
        ({0}, 1), ({1}, 6), ({2}, 2), ({3}, 3), ({4}, 8),
        ({0, 2}, 4), ({0, 3}, 5), ({1, 3}, 7), ({2, 3}, 7),
    ]
    passkeys_true = {frozenset(key): intent_i for key, intent_i in passkeys_true}

    pseudo_intents_true = [({4}, 8), ({1}, 6), ({2, 3}, 7), ({0, 1, 2}, 8)]
    pseudo_intents_true = {frozenset(pi): intent_i for pi, intent_i in pseudo_intents_true}

    proper_premises_true = [({1}, 6), ({4}, 8), ({0, 1}, 8), ({2, 3}, 7), ({0, 2, 3}, 8)]
    proper_premises_true = {frozenset(pp): intent_i for pp, intent_i in proper_premises_true}

    parents_ordering_true = [set(), {0}, {0}, {0}, {1, 2}, {1, 3}, {2}, {3, 6}, {4, 5, 7}]
    transitive_parents = [set(), {0}, {0}, {0}, {0, 1, 2}, {0, 1, 3}, {0, 2}, {0, 2, 3, 6}, {0, 1, 2, 3, 4, 5, 6, 7}]
    n_trans_parents = sum(len(tpars) for tpars in transitive_parents)
    linearity_true = linearity_index(n_trans_parents, len(intents_true))
    distributivity_true = distributivity_index(
        list(isets2bas(intents_true, K.shape[1])),
        list(isets2bas(parents_ordering_true, len(intents_true))),
        n_trans_parents
    )

    explore_data_true = dict(
        intents=intents_true, keys=keys_true, passkeys=passkeys_true,
        pseudo_intents=pseudo_intents_true, proper_premises=proper_premises_true,
        intents_ordering=parents_ordering_true, linearity=linearity_true, distributivity=distributivity_true
    )

    explored_data = explore_data(K)
    assert explored_data['intents'] == intents_true
    assert explored_data['keys'] == keys_true
    assert explored_data['passkeys'] == passkeys_true
    assert explored_data['pseudo_intents'] == pseudo_intents_true
    assert explored_data['proper_premises'] == proper_premises_true
    assert explored_data['intents_ordering'] == parents_ordering_true
    assert explored_data['linearity'] == linearity_true
    assert explored_data['distributivity'] == distributivity_true

    assert explored_data == explore_data_true



import typing
from dataclasses import dataclass
from itertools import combinations, islice
from string import ascii_uppercase
from typing import Iterable, Iterator, FrozenSet, BinaryIO, Generator, Union, TextIO, Optional, get_args, Sequence
import numpy.typing as npt
import numpy as np
import pandas as pd
from functools import reduce

from bitarray import frozenbitarray as fbarray, bitarray
from bitarray.util import zeros as bazeros

ItemsetContextType = list[set[int]]  # `j in ctx[i]` shows whether i-th object shares j-th attribute
NamedItemsetContextType = tuple[ItemsetContextType, list[str], list[str]]  # Itemsets + object names + attribute names
BitarrayContextType = list[bitarray]  # `ctx[i][j]` shows whether i-th object shares j-th attribute
NamedBitarrayContextType = tuple[BitarrayContextType, list[str], list[str]]  # bitarrays + object names + attr. names
BoolContextType = list[list[bool]]  # `ctx[i][j]` shows whether i-th object shares j-th attribute
NamedBoolContextType = tuple[BoolContextType, list[str], list[str]]  # bools context + object names + attribute names
DictContextType = dict[str, set[str]]  # `m in ctx[g]` shows whether object g shares attribute m
PandasContextType = pd.DataFrame  # pandas dataframe where row indices are objects and columns indices are attrs.

ContextType = Union[
    PandasContextType,
    ItemsetContextType, NamedItemsetContextType,
    BitarrayContextType, NamedBitarrayContextType,
    BoolContextType, NamedBoolContextType,
    DictContextType
]


@dataclass
class UnknownContextTypeError(TypeError):
    submitted_type: type

    def __str__(self) -> str:
        return f"Received context is of unknown type: {self.submitted_type}. " +\
                f"Supported types are: \n" +\
                '\n'.join([f"* {t}" for t in get_args(ContextType)])


##########################
# Basic type conversions #
##########################
def np2bas(X: npt.ArrayLike) -> list[fbarray]:
    """Convert numpy boolean matrix to the list of bitarrays (one per row of the matrix)"""
    return [fbarray(x) for x in X.tolist()]


def bas2np(barrays: Iterable[fbarray]) -> npt.ArrayLike:
    """Convert the list of bitarrays to numpy boolean matrix (bitarrays become rows in of the matrix)"""
    return np.vstack([ba.tolist() for ba in barrays]).astype(np.bool_)


def isets2bas(itemsets: Iterable[Iterable[int]], length: int) -> Iterator[fbarray]:
    """Convert the list of lists of indices of 'True' elements to bitarrays of given length

    Examples
    --------
    isets2bas([ [0, 1], [1,3,4] ], 5) --> [bitarray('01000'), bitarray('01011')]
    """
    for iset in itemsets:
        bar = bazeros(length)
        for m in iset:
            bar[m] = True
        yield fbarray(bar)


def bas2isets(bitarrays: Iterable[fbarray]) -> Iterator[FrozenSet[int]]:
    """Convert the list of bitarrays to the list of lists of indices of 'True' elements of the bitarrays

    Examples
    --------
    bas2isets([ bitarray('01000'), bitarray('01011') ]) --> [ [0,1], [1,3,4] ]
    """
    for bar in bitarrays:
        yield frozenset(bar.itersearch(True))


def to_named_itemsets(data: ContextType) -> NamedItemsetContextType:
    """Convert the context defined by `data` into the named itemset format

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats:
         pd.DataFrame (with bool values);
         dictionary with object names as keys and named object descriptions as values;
         itemsets where every object description is represented with a set of indices of described attributes;
         bitarrays where every object description is represented with a bitarray where every value corresponds to an attribute;
         list of bools where every object description is represented with a list of bools where every value corresponds to an attribute;

         For itemset, bitarrays, and list of bools types of contexts,
         one can also provide the names of objects and attributes (as shown in the examples).

    Return
    ------
    itemsets: list[set[int]]
        list of sets of indices of True-valued columns
    object_names: list[str]
        names of objects (i.e. rows) in the data.
        If the names are not specified by the data, returns "object_1", "object_2", ...
    attribute_names: list[str]
        names of attributes (i.e. columns) in the data
        If the names are not specified by the data, returns "attribute_1", "attribute_2", ...

    Examples
    --------
    to_named_itemsets( pd.DataFrame({'a': [False, True], 'b': [True, True]}) )
        --> ([{1}, {0, 1}], ['0', '1'], ['a', 'b'])

    to_named_itemsets( {'row1': ['b'], 'row2': ['a', 'b']} )
        --> ([{1}, {0, 1}], ['row1', 'row2'], ['a', 'b'])

    to_named_itemsets( [[1], [0,1]] )
        --> ([{1}, {0, 1}], ['object_0', 'object_1'], ['attribute_0', 'attribute_1'])

    to_named_itemsets( ([[1], [0,1]], ['g1', 'g2'], ['m1', 'm2']) )
        --> ([{1}, {0, 1}], ['g1', 'g2'], ['m1', 'm2'])
    """
    if len(data) == 0:
        return [], [], []

    context_type = identify_supported_context_type(data)
    if context_type is None:
        raise UnknownContextTypeError(type(data))

    if context_type == NamedItemsetContextType:
        return data

    if context_type == PandasContextType:
        return list(map(set, bas2isets(np2bas(data.values)))), list(map(str, data.index)), list(map(str, data.columns))

    if context_type == DictContextType:
        objects = sorted(data.keys())
        attributes = sorted(reduce(set.union, data.values(), set()))
        attrs_idx_map = {attr: attr_i for attr_i, attr in enumerate(attributes)}

        itemsets = [[attrs_idx_map[attr] for attr in data[obj]] for obj in objects]
        return list(map(set, itemsets)), list(map(str, objects)), list(map(str, attributes))

    # Possible data types: Itemsets or (Named) Bitarrays or (Named) Bools
    is_named = context_type in {NamedItemsetContextType, NamedBitarrayContextType, NamedBoolContextType}
    crosses_data = data[0] if is_named else data
    objects = data[1] if is_named else (f'object_{i}' for i in range(len(crosses_data)))

    n_attrs = len(reduce(set.__or__, map(set, crosses_data))) if context_type == ItemsetContextType else len(crosses_data[0])
    attributes = data[2] if is_named else (f'attribute_{j}' for j in range(n_attrs))
    objects, attributes = list(map(str, objects)), list(map(str, attributes))

    if context_type == ItemsetContextType:
        return list(map(set, crosses_data)), objects, attributes

    # Possible data types: (Named) Bitarrays or (Named) Bools
    return list(map(set, bas2isets(map(bitarray, crosses_data)))), objects, attributes


def to_itemsets(data: ContextType) -> ItemsetContextType:
    """Convert the context defined by `data` into the itemset format

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats.
        The list of the supported formats is provided in the Docstring of `io.to_named_itemsets(...)` function.

    Return
    ------
    itemsets: list[set[int]]
        list of sets of indices of True-valued columns

    Examples
    --------
    to_itemsets( pd.DataFrame({'a': [False, True], 'b': [True, True]}) )
        --> [{1}, {0, 1}]

    to_itemsets( {'row1': ['b'], 'row2': ['a', 'b']} )
        --> [{1}, {0, 1}]

    to_itemsets( [[1], [0, 1]] )
        --> [{1}, {0, 1}]
    """
    return to_named_itemsets(data)[0]


def to_dictionary(data: ContextType) -> DictContextType:
    """Convert the context defined by `data` into the dictionary format

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats.
        The list of the supported formats is provided in the Docstring of `io.to_named_itemsets(...)` function.

    Return
    ------
    dictionary: dict[str, set[str]]
        dictionary of type 'object_name' -> set of names of attributes describing object

    Examples
    --------
    to_dictionary( pd.DataFrame({'a': [False, True], 'b': [True, True]}) )
        -->  {'0': {'b'}, '1': {'a', 'b'}}

    to_dictionary( [bitarray('01'), bitarray('11')] )
        --> {'object_0': {'attribute_1'}, 'object_1': {'attribute_0', 'attribute_1'}}

    to_dictionary( [[1], [0,1]] )
        --> {'object_0': {'attribute_1'}, 'object_1': {'attribute_0', 'attribute_1'}}
    """
    itemsets, objects, attributes = to_named_itemsets(data)
    return {obj: {attributes[attr_i] for attr_i in itemset} for obj, itemset in zip(objects, itemsets)}


def to_named_bitarrays(data: ContextType) -> NamedBitarrayContextType:
    """Convert the context defined by `data` into the named bitarrays format

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats.
        The list of the supported formats is provided in the Docstring of `io.to_named_itemsets(...)` function.

    Return
    ------
    bitarrays: list[bitarray]
        list of frozen bitarrays representing descriptions of objects. For i-th object and j-th attribute,
        `bitarrays[i][j] == True` means that i-th object is described by j-th attribute.
    object_names: list[str]
        names of objects (i.e. rows) in the data.
        If the names are not specified by the data, returns "object_1", "object_2", ...
    attribute_names: list[str]
        names of attributes (i.e. columns) in the data
        If the names are not specified by the data, returns "attribute_1", "attribute_2", ...

    Examples
    --------
    to_named_bitarrays( pd.DataFrame({'a': [False, True], 'b': [True, True]}) )
        --> ([bitarray('01'), bitarray('11')], ['0', '1'], ['a', 'b'])

    to_named_bitarrays( {'row1': ['b'], 'row2': ['a','b']} )
        -->  ([bitarray('01'), bitarray('11')], ['row1', 'row2'], ['a', 'b'])

    to_named_bitarrays( [[1], [0, 1]] )
        --> ([bitarray('01'), bitarray('11')], ['object_0', 'object_1'], ['attribute_0', 'attribute_1'])
    """
    itemsets, objects, attributes = to_named_itemsets(data)
    return list(map(bitarray, isets2bas(itemsets, len(attributes)))), objects, attributes


def to_bitarrays(data: ContextType) -> BitarrayContextType:
    """Convert the context defined by `data` into the named bitarrays format

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats.
        The list of the supported formats is provided in the Docstring of `io.to_named_itemsets(...)` function.

    Return
    ------
    bitarrays: list[bitarray]
        list of frozen bitarrays representing descriptions of objects. For i-th object and j-th attribute,
        `bitarrays[i][j] == True` means that i-th object is described by j-th attribute.

    Examples
    --------
    to_bitarrays( pd.DataFrame({'a': [False, True], 'b': [True, True]}) )
        --> [bitarray('01'), bitarray('11')]

    to_bitarrays( {'row1': ['b'], 'row2': ['a','b']} )
        --> [bitarray('01'), bitarray('11')]

    to_bitarrays( [[1], [0, 1]] )
        --> [bitarray('01'), bitarray('11')]
    """
    return to_named_bitarrays(data)[0]


def to_named_bools(data: ContextType) -> NamedBoolContextType:
    """Convert the context defined by `data` into the named bools format

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats.
        The list of the supported formats is provided in the Docstring of `io.to_named_itemsets(...)` function.

    Return
    ------
    bools: list[list[bool]
        list of lists of bools representing descriptions of objects. For i-th object and j-th attribute,
        `bools[i][j] == True` means that i-th object is described by j-th attribute.
    object_names: list[str]
        names of objects (i.e. rows) in the data.
        If the names are not specified by the data, returns "object_1", "object_2", ...
    attribute_names: list[str]
        names of attributes (i.e. columns) in the data
        If the names are not specified by the data, returns "attribute_1", "attribute_2", ...

    Examples
    --------
    to_named_bools( pd.DataFrame({'a': [False, True], 'b': [True, True]}) )
        --> ([[False, True], [True, True]], ['0', '1'], ['a', 'b'])

    to_named_bools( {'row1': ['b'], 'row2': ['a','b']} )
        -->  ([[False, True], [True, True]], ['row1', 'row2'], ['a', 'b'])

    to_named_bools( [[1], [0, 1]] )
        --> ([[False, True], [True, True]], ['object_0', 'object_1'], ['attribute_0', 'attribute_1'])
    """
    itemsets, objects, attributes = to_named_itemsets(data)
    bools = [[i in itemset for i in range(len(attributes))] for itemset in itemsets]
    return bools, objects, attributes


def to_bools(data: ContextType) -> BoolContextType:
    """Convert the context defined by `data` into the named bools format

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats.
        The list of the supported formats is provided in the Docstring of `io.to_named_itemsets(...)` function.

    Return
    ------
    bools: list[list[bool]
        list of lists of bools representing descriptions of objects. For i-th object and j-th attribute,
        `bools[i][j] == True` means that i-th object is described by j-th attribute.

    Examples
    --------
    to_bools( pd.DataFrame({'a': [False, True], 'b': [True, True]}) )
        --> [[False, True], [True, True]]

    to_bools( {'row1': ['b'], 'row2': ['a','b']} )
        -->  [[False, True], [True, True]]

    to_bools( [[1], [0, 1]] )
        --> [[False, True], [True, True]]
    """
    return to_named_bools(data)[0]


def to_pandas(data: ContextType) -> pd.DataFrame:
    """Convert the context defined by `data` into pandas DataFrame format

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats.
        The list of the supported formats is provided in the Docstring of `io.to_named_itemsets(...)` function.

    Return
    ------
    dataframe: pd.DataFrame
        Binary pandas.DataFrame representing the `data`

    Examples
    --------
    to_pandas( {'row1': ['b'], 'row2': ['a','b']} )
        --> >           a     b
            > row1  False  True
            > row2   True  True

    to_pandas( [[1], [0, 1]] )
        --> >           attribute_0  attribute_1
            > object_0        False         True
            > object_1         True         True

    to_pandas( [bitarray('01'), bitarray('11')] )
        --> >           attribute_0  attribute_1
            > object_0        False         True
            > object_1         True         True
    """
    itemsets, objects, attributes = to_named_itemsets(data)
    df = pd.DataFrame(False, index=objects, columns=attributes).fillna(False)
    for obj_i, itemset in enumerate(itemsets):
        df.iloc[obj_i, list(itemset)] = True
    return df
    

def transpose_context(data: ContextType) -> ContextType:
    """Switch places among objects and attributes and transpose their relation

    Parameters
    ----------
    data: ContextType
        Binary formal context in one of the supported formats.
        The list of the supported formats is provided in the Docstring of `io.to_named_itemsets(...)` function.

    Return
    ------
    transposed_data: ContextType
        Transposed context. The format of transposed data is the same as the format of the original data
    """
    if len(data) == 0:
        return data

    data_type = identify_supported_context_type(data)
    if data is None:
        raise UnknownContextTypeError(type(data))

    if data_type == PandasContextType:
        return data.T

    if data_type == DictContextType:
        transposed = {}
        for obj, description in data.items():
            for attr in description:
                if attr not in transposed:
                    transposed[attr] = []
                transposed[attr].append(obj)
        return {k: set(v) for k, v in transposed.items()}

    is_named = data_type in {NamedItemsetContextType, NamedBitarrayContextType, NamedBoolContextType}
    crosses_data = data[0] if is_named else data
    if data_type in {NamedItemsetContextType, ItemsetContextType}:
        n_attrs = len(data[2]) if is_named else len(reduce(set.__or__, crosses_data, set()))
        transposed = [[] for _ in range(n_attrs)]
        for obj_i, itemset in enumerate(crosses_data):
            for attr_i in itemset:
                transposed[attr_i].append(obj_i)
        transposed = list(map(set, transposed))

    if data_type in {NamedBitarrayContextType, BitarrayContextType}:
        n_objs, n_attrs = len(crosses_data), len(crosses_data[0])
        transposed = [bazeros(n_objs) for _ in range(n_attrs)]
        for obj_i, ba in enumerate(crosses_data):
            for attr_i in ba.itersearch(True):
                transposed[attr_i][obj_i] = True
        transposed = list(map(bitarray, transposed))

    if data_type in {NamedBoolContextType, BoolContextType}:
        n_objs, n_attrs = len(crosses_data), len(crosses_data[0])
        transposed = [[False] * n_objs for _ in range(n_attrs)]
        for obj_i, bools in enumerate(crosses_data):
            for attr_i, b in enumerate(bools):
                if b:
                    transposed[attr_i][obj_i] = True

    if is_named:
        return transposed, data[2], data[1]
    return transposed


def identify_supported_context_type(context: ContextType) -> Optional[typing.Type]:
    """Return the supported TypeName of the provided `context`. Return None if the context format is not supported"""
    is_pandas = isinstance(context, pd.DataFrame)
    is_dict = isinstance(context, dict) \
              and all(isinstance(k, str) for k in context.keys()) \
              and all(all(isinstance(v, str) for v in vals) for vals in context.values())
    is_named = isinstance(context, Sequence) and len(context) == 3 and all(
        all(isinstance(v, str) for v in names) for names in context[1:])
    crosses_data = context[0] if is_named else context
    is_sequence = isinstance(crosses_data, Sequence)
    is_bitarrays = is_sequence and isinstance(crosses_data[0], bitarray)
    is_bools = is_sequence and isinstance(crosses_data, (list, tuple)) \
               and all(all(isinstance(v, bool) for v in vals) for vals in crosses_data)
    is_itemsets = is_sequence and not is_bitarrays and not is_bools and isinstance(list(crosses_data[0])[0], int)

    dtypes_selector = [
        (is_pandas, PandasContextType), (is_dict, DictContextType),
        (not is_named and is_itemsets, ItemsetContextType), (is_named and is_itemsets, NamedItemsetContextType),
        (not is_named and is_bitarrays, BitarrayContextType), (is_named and is_bitarrays, NamedBitarrayContextType),
        (not is_named and is_bools, BoolContextType), (is_named and is_bools, NamedBoolContextType),
    ]
    if sum(condition for condition, dtype in dtypes_selector) != 1:
        return None

    return next(dtype for condition, dtype in dtypes_selector if condition)


def verbalise(description: Union[bitarray, Iterable[int]], names: list[str]) -> Iterable[str]:
    """Convert every index i (or every True i-th element of a bitarray) into a human-readable string `names[i]`"""
    if isinstance(description, bitarray):
        return {names[i] for i in description.itersearch(True)}

    if isinstance(description, Generator):
        return (names[i] for i in description)

    return type(description)([names[i] for i in description])


def to_absolute_number(percentage: Union[int, float], total_size: int) -> int:
    """Convert a float percentage into the absolute number of elements w.r.t. `total_size`. Do nothing with integers"""
    if isinstance(percentage, int):
        return percentage
    return int(percentage * total_size)


###########################
# Save and load functions #
###########################

def load_balist(file: BinaryIO) -> Iterator[bitarray]:
    """Load the list of bitarrays from binary file"""
    basize = b''
    while True:
        data = file.read(1)
        if data == b'n':
            break
        basize += data
    basize = int(basize.decode(encoding='utf-8'))
    basize_bytes = len(bazeros(basize).tobytes())

    while True:
        data = file.read(basize_bytes)
        if data == b'':
            break

        ba = bitarray()
        ba.frombytes(data)
        yield ba[:basize]


def save_balist(file: BinaryIO, bitarrays: list[bitarray]):
    """Save the list of bitarrays to the binary file (proposed file extension '.bal')"""
    basize = len(bitarrays[0])
    assert all(len(ba) == basize for ba in bitarrays), "All bitarrays should be of the same size"

    file.write(str(basize).encode(encoding='utf-8'))
    file.write(b'n')

    for ba in bitarrays:
        file.write(ba.tobytes())


def read_cxt(file: Union[TextIO, str]) -> PandasContextType:
    """Read the file (or string) that describes a formal context in Burmeister format"""
    data = file.read() if not isinstance(file, str) else file

    _, ns, data = data.split('\n\n')
    n_objs, n_attrs = [int(x) for x in ns.split('\n')]

    data = data.strip().split('\n')
    objects, data = data[:n_objs], data[n_objs:]
    attributes, data = data[:n_attrs], data[n_attrs:]

    crosses = [[c == 'X' for c in line] for line in data]
    return to_pandas((crosses, objects, attributes))


def write_cxt(context: ContextType, file: TextIO = None) -> str:
    """Return the formal `context` represented with Burmeister format. Save the string to a `file` if it is provided"""
    crosses, objects, attributes = to_named_itemsets(context)

    file_data = 'B\n\n'
    file_data += f"{len(objects)}\n{len(attributes)}\n"
    file_data += '\n'
    file_data += '\n'.join(objects) + '\n'
    file_data += '\n'.join(attributes) + '\n'

    file_data += '\n'.join([''.join(
        ['X' if idx in crosses_line else '.' for idx in range(len(attributes))])
        for crosses_line in crosses
    ]) + '\n'

    if file:
        file.write(file_data)
    return file_data


def from_fca_repo(context_name: str) -> PandasContextType:
    """Download a formal context from the (git) repository of contexts.

    Go to 'https://fcarepository.org' for more information on the repo.
    """
    import urllib.request

    context_name = context_name + '.cxt' if not context_name.endswith('.cxt') else context_name

    url = f"https://github.com/fcatools/contexts/raw/main/contexts/{context_name}"
    context_data = urllib.request.urlopen(url).read().decode("utf-8")
    return read_cxt(context_data)


def to_mermaid_diagram(node_labels: list[str], neighbours: list[list[int]]) -> str:
    """Create a mermaid flowchart code. The code can be visualised via https://mermaid.live/ or via GitHub markdown"""
    nodes_symbols = (''.join(symbols) for symbols_len in range(1, len(node_labels)+1)
                     for symbols in combinations(ascii_uppercase, symbols_len))
    nodes_symbols = list(islice(nodes_symbols, len(node_labels)))

    nodes_lines = [f'{symbol}["{label}"];' for symbol, label in zip(nodes_symbols, node_labels)]
    edges_lines = [f"{node} --- {nodes_symbols[neighbour]};"
                   for node, neighbours_indices in zip(nodes_symbols, neighbours) for neighbour in neighbours_indices]

    mermaid_lines = ['flowchart TD'] + nodes_lines + [''] + edges_lines
    return '\n'.join(mermaid_lines)



import numpy as np
import pandas as pd
from bitarray import frozenbitarray as fbarray
import os

from caspailleur import io


def test_np_ba_conversions():
    X = np.array([[True, True, False], [False, False, True]])
    bars_true = [fbarray(x) for x in X.tolist()]
    bars = io.np2bas(X)
    assert bars == bars_true

    X_reconstr = io.bas2np(bars)
    assert (X == X_reconstr).all()

    attr_exts_true = [fbarray([True, False]), fbarray([True, False]), fbarray([False, True])]
    assert io.np2bas(X.T) == attr_exts_true


def test_iset_ba_conversions():
    iset, l = {0, 1}, 5
    ba = fbarray([True, True, False, False, False])

    assert list(io.isets2bas([iset], l)) == [ba]
    assert list(io.bas2isets([ba])) == [iset]
    assert list(io.bas2isets(io.isets2bas([iset], l))) == [iset]


def test_conversion_pipeline():
    X = np.array([[True, True, False], [False, False, True]])
    itemsets_true = [{0, 1}, {2}]
    itemsets = list(io.bas2isets(io.np2bas(X)))
    assert itemsets == itemsets_true

    X_reconstr = io.bas2np(io.isets2bas(itemsets, 3))
    assert (X == X_reconstr).all()


def test_save_load_barrays():
    bitarrays = [fbarray([True, False]), fbarray([True, False]), fbarray([False, True])]
    with open('tst.bal', 'wb') as file:
        io.save_balist(file, bitarrays)
    with open('tst.bal', 'rb') as file:
        bitarrays_load = list(io.load_balist(file))

    os.remove('tst.bal')

    assert bitarrays_load == bitarrays


def test_context_conversions():
    datas = dict(
        pandas=pd.DataFrame([[True, True, False], [False, False, True]],
                            index=['g1', 'g2'], columns=['m1', 'm2', 'm3']),
        itemset=[{0, 1}, {2}],
        named_itemset=([{0, 1}, {2}], ['g1', 'g2'], ['m1', 'm2', 'm3']),
        bitarray=[fbarray('110'), fbarray('001')],
        named_bitarray=([fbarray('110'), fbarray('001')], ['g1', 'g2'], ['m1', 'm2', 'm3']),
        bool=[[True, True, False], [False, False, True]],
        named_bool=([[True, True, False], [False, False, True]], ['g1', 'g2'], ['m1', 'm2', 'm3']),
        dict={'g1': {'m1', 'm2'}, 'g2': {'m3'}}
    )
    objects_map = {'g1': 'object_0', 'g2': 'object_1'}
    attributes_map = {'m1': 'attribute_0', 'm2': 'attribute_1', 'm3': 'attribute_2'}

    funcs = dict(
        pandas=io.to_pandas,
        itemset=io.to_itemsets,
        named_itemset=io.to_named_itemsets,
        bitarray=io.to_bitarrays,
        named_bitarray=io.to_named_bitarrays,
        bool=io.to_bools,
        named_bool=io.to_named_bools,
        dict=io.to_dictionary
    )
    named_types = {'pandas', 'named_itemset', 'named_bitarray', 'named_bool', 'dict'}

    assert datas.keys() == funcs.keys()
    for input_type in datas.keys():
        for output_type in [
            'itemset',
            'named_itemset',
            'bitarray',
            'named_bitarray',
            'bool',
            'named_bool',
            'dict'
        ]:  # datas.keys():
            input_data, output_data = datas[input_type], datas[output_type]

            # if the transformation will forget the objects/attributes names
            if (input_type not in named_types) and (output_type in named_types):
                if output_type == 'pandas':
                    output_data = output_data.rename(index=objects_map, columns=attributes_map)
                elif output_type == 'dict':
                    output_data = {objects_map[obj]: {attributes_map[attr] for attr in description}
                                  for obj, description in output_data.items()}
                else:
                    output_data = output_data[0],\
                        [objects_map[g] for g in output_data[1]], [attributes_map[m] for m in output_data[2]]

            print(input_type, output_type)
            test_data = funcs[output_type](input_data)
            assert test_data == output_data, f"Having problems transforming {input_type} context into {output_type}"


def test_transpose_context():
    datas = dict(
        pandas=pd.DataFrame([[True, True, False], [False, False, True]],
                            index=['g1', 'g2'], columns=['m1', 'm2', 'm3']),
        itemset=[{0, 1}, {2}],
        named_itemset=([{0, 1}, {2}], ['g1', 'g2'], ['m1', 'm2', 'm3']),
        bitarray=[fbarray('110'), fbarray('001')],
        named_bitarray=([fbarray('110'), fbarray('001')], ['g1', 'g2'], ['m1', 'm2', 'm3']),
        bool=[[True, True, False], [False, False, True]],
        named_bool=([[True, True, False], [False, False, True]], ['g1', 'g2'], ['m1', 'm2', 'm3']),
        dict={'g1': {'m1', 'm2'}, 'g2': {'m3'}}
    )
    datas_transposed = dict(
        pandas=pd.DataFrame([[True, False], [True, False], [False, True]],
                            index=['m1', 'm2', 'm3'], columns=['g1', 'g2']),
        itemset=[{0}, {0}, {1}],
        named_itemset=([{0}, {0}, {1}], ['m1', 'm2', 'm3'], ['g1', 'g2']),
        bitarray=[fbarray('10'), fbarray('10'), fbarray('01')],
        named_bitarray=([fbarray('10'), fbarray('10'), fbarray('01')], ['m1', 'm2', 'm3'], ['g1', 'g2']),
        bool=[[True, False], [True, False], [False, True]],
        named_bool=([[True, False], [True, False], [False, True]], ['m1', 'm2', 'm3'], ['g1', 'g2']),
        dict={'m1': {'g1'}, 'm2': {'g1'}, 'm3': {'g2'}}
    )

    for data_type, data in datas.items():
        transposed = io.transpose_context(data)
        if data_type == 'pandas':
            assert (transposed == datas_transposed[data_type]).all(None), "Problem transposing pd.DataFrame"
            assert (io.transpose_context(transposed) == data).all(None), "Problem transposing pd.DataFrame twice"
            continue

        assert transposed == datas_transposed[data_type], f"Problem transposing {data_type} context"
        assert io.transpose_context(transposed) == data, f"Problem transposing {data_type} context twice"


def test_read_write_cxt():
    data_string = '\n'.join([
        'B', '', '3', '3', '',
        'Consulting', 'Planning', 'Assembly and installation',
        'Furniture', 'Computers', 'Copy machines',
        'XXX', 'XX.', 'XXX', ''
    ])

    data_df = pd.DataFrame(
        [[True, True, True], [True, True, False], [True, True, True]],
        index=['Consulting', 'Planning', 'Assembly and installation'],
        columns=['Furniture', 'Computers', 'Copy machines']
    )
    assert (io.read_cxt(data_string) == data_df).all(None)
    assert io.write_cxt(data_df) == data_string

    with open('test_file.cxt', 'w') as file:
        io.write_cxt(data_df, file)
    with open('test_file.cxt', 'r') as file:
        data_read = io.read_cxt(file)
    assert (data_read == data_df).all(None)
    os.remove('test_file.cxt')


def test_from_fca_repo():
    context_name = 'planets_en'
    assert (io.from_fca_repo(context_name) == io.from_fca_repo(context_name+'.cxt')).all(None)

    context_df = pd.DataFrame([
        [True, False, False, True, False, False, True],  # X..X..X
        [True, False, False, True, False, False, True],  # X..X..X
        [True, False, False, True, False, True, False],  # X..X.X.
        [True, False, False, True, False, True, False],  # X..X.X.
        [False, False, True, False, True, True, False],  # ..X.XX.

        [False, False, True, False, True, True, False],  # ..X.XX.
        [False, True, False, False, True, True, False],  # .X..XX.
        [False, True, False, False, True, True, False],  # .X..XX.
        [True, False, False, False, True, True, False],  # X...XX.
    ],
        index=['Merkur', 'Venus', 'Earth', 'Mars', 'Jupiter', 'Saturn', 'Uranus', 'Neptune', 'Pluto'],
        columns=['Small', 'Medium', 'Large', 'Near', 'Distant', 'Moon', 'No moon']
    )

    assert (io.from_fca_repo(context_name) == context_df).all(None)


def test_to_mermaid_diagram():
    nodes = ['Top', 'Left Top', 'Left Bottom', 'Right', 'Bottom']
    neighbours = [[1, 3], [2], [4], [4], []]
    diagram_text = '\n'.join([
        'flowchart TD',
        'A["Top"];', 'B["Left Top"];', 'C["Left Bottom"];', 'D["Right"];', 'E["Bottom"];', '',
        'A --- B;', 'A --- D;', 'B --- C;', 'C --- E;', 'D --- E;'
    ])

    txt = io.to_mermaid_diagram(nodes, neighbours)
    assert txt == diagram_text



from typing import Iterator, Iterable, Union
from collections import deque
from bitarray import frozenbitarray as fbarray, bitarray
from tqdm.auto import tqdm

from caspailleur.order import check_topologically_sorted
from caspailleur.base_functions import extension


############################
# Indices for descriptions #
############################
def delta_stability_by_extents(extents: list[fbarray]) -> Iterator[int]:
    """Compute the delta stability index: the difference in supports of an extent and its maximal smaller neighbour"""
    assert check_topologically_sorted(extents)

    for i, extent in enumerate(extents):
        if i == 0:
            yield extent.count()
            continue

        for smaller_extent in extents[i - 1::-1]:
            if smaller_extent & extent != smaller_extent:
                continue

            yield (extent & ~smaller_extent).count()
            break
        else:  # no break, i.e. no child extent found
            yield extent.count()


def delta_stability_by_description(
        description: Union[Iterable[int], bitarray], crosses_per_columns: list[fbarray], extent: fbarray = None
) -> int:
    description = set(description.itersearch(True)) if isinstance(description, bitarray) else set(description)

    extent = extension(description, crosses_per_columns) if extent is None else extent
    out_attrs = [(m_i, m_ext) for m_i, m_ext in enumerate(crosses_per_columns) if m_i not in description]
    if not out_attrs:
        return extent.count()

    max_sub_support = max((extent & m_ext).count() for m_i, m_ext in out_attrs)
    delta_stab = extent.count() - max_sub_support
    return delta_stab


def support_by_description(
        description: Union[Iterable[int], bitarray], crosses_per_columns: list[fbarray], extent: fbarray = None
) -> int:
    extent = extension(description, crosses_per_columns) if extent is None else extent
    return extent.count()


##############################
# Lattice complexity indices #
##############################

def linearity_index(n_trans_parents: int, n_elements: int, include_top_bottom: bool = True) -> float:
    """Compute linearity index: the percentage of comparable pairs of elements

    Parameters
    ----------
    n_trans_parents:
        Number of transitive parents, i.e. the cardinality of transitive order relation
    n_elements:
        Number of elements
    include_top_bottom:
        A flag, whether to include top/bottom nodes in the computations

    Returns
    -------
    float:
        Value of the index (from 0 to 1)
    """
    n_comparable = n_trans_parents
    n_pairs = n_elements * (n_elements - 1) // 2

    if not include_top_bottom:
        # for (n_elems - 2) elements between top and bottom, drop pairs with the top,
        # for bottom element drop all (n_elems - 1) relations
        n_comparable -= 2*n_elements - 3
        n_pairs -= 2*n_elements - 3

    if n_comparable > n_pairs:
        raise ValueError('Linearity index is computed in a wrong way (There should be problem with the code)')

    return n_comparable / n_pairs if n_pairs else 0


def distributivity_index(
        intents: list[fbarray], parents: list[fbarray], n_trans_parents: int,
        include_top_bottom: bool = True, use_tqdm: bool = False
) -> float:
    """Compute distributivity index: the percentage of pairs intents that whose union is in an intent

    Parameters
    ----------
    intents:
        The list of intents
    parents:
        Parents relation represented by list of bitarrays: each intent_i holds indices of parents of intent_i
    n_trans_parents:
        Number of transitive parents, i.e. the cardinality of the transitive relation
    include_top_bottom:
        A flag, whether to include top/bottom nodes in the computations
    use_tqdm:
        A flag, whether to visualise the computation process via tqdm progressbar

    Returns
    -------
    float:
        Value of the index (from 0 to 1)
    """
    assert check_topologically_sorted(intents), 'The `intents` list should be topologically sorted by ascending order'

    n_distr = n_trans_parents

    for intent_idx, intent in tqdm(
            enumerate(intents), total=len(intents),
            disable=not use_tqdm, desc='enumerate intents'
    ):
        distr_ancestors = deque([
            (mother, father)
            for mother in parents[intent_idx].itersearch(True) for father in parents[intent_idx].itersearch(True)
            if mother < father and intents[mother] | intents[father] == intent
        ])

        visited_pairs = set()
        while distr_ancestors:
            pair = distr_ancestors.popleft()
            if pair in visited_pairs or pair[::-1] in visited_pairs:
                continue
            visited_pairs.add(pair)

            n_distr += 1

            mother, father = pair
            mother_intent, father_intent = intents[mother], intents[father]

            distr_ancestors.extend([
                (mother, gfather) for gfather in parents[father].itersearch(True)
                if mother_intent | intents[gfather] == intent
            ])
            distr_ancestors.extend([
                (gmother, father) for gmother in parents[mother].itersearch(True)
                if intents[gmother] | father_intent == intent
            ])

    n_intents = len(intents)
    n_pairs = n_intents * (n_intents - 1) // 2
    if not include_top_bottom:
        n_distr -= 2 * n_intents - 3  # dropping (n-1) rels for top intent and 1 rel. for (n-2) intents
        n_pairs -= 2 * n_intents - 3

    if n_distr > n_pairs:
        raise ValueError('Distributivity index is computed in a wrong way (There should be problem with the code)')

    return n_distr / n_pairs if n_pairs else 0



from caspailleur import definitions as defs
from caspailleur.base_functions import powerset


def test_is_closed():
    crosses_per_columns = [
        {1, 2},     # a
        {3, 4},     # b
        {2, 3, 4},  # c
        {1, 4},     # d
        set(),      # e
    ]
    attrs = 'abcde'
    closed_sets_true = {'bc', 'bcd', 'abcde', 'a', 'c', 'd', 'ac', 'ad', ''}

    powset = powerset(range(len(attrs)))
    closed_sets = {B for B in powset if defs.is_closed(B, crosses_per_columns)}
    closed_sets_verb = {''.join([attrs[m] for m in sorted(B)]) for B in closed_sets}
    assert closed_sets_verb == closed_sets_true


def test_is_pseudo_intent():
    crosses_per_columns = [
        {1, 2},     # a
        {3, 4},     # b
        {2, 3, 4},  # c
        {1, 4},     # d
        set(),      # e
    ]
    attrs = 'abcde'
    pseudo_intents_true = {'b', 'e', 'cd', 'abc'}

    powset = list(powerset(range(len(attrs))))

    pseudo_intents = []
    for B in powset:
        if defs.is_pseudo_intent(B, crosses_per_columns, pseudo_intents):
            pseudo_intents.append(B)
    pseudo_intents_verb = {''.join([attrs[m] for m in sorted(B)]) for B in pseudo_intents}
    assert pseudo_intents_verb == pseudo_intents_true

    pseudo_intents = {B for B in powset if defs.is_pseudo_intent(B, crosses_per_columns)}
    pseudo_intents_verb = {''.join([attrs[m] for m in sorted(B)]) for B in pseudo_intents}
    assert pseudo_intents_verb == pseudo_intents_true


def test_is_proper_premise():
    crosses_per_columns = [
        {1, 2},  # a
        {3, 4},  # b
        {2, 3, 4},  # c
        {1, 4},  # d
        set(),  # e
    ]
    attrs = 'abcde'
    pprems_true = {'b', 'e', 'cd', 'ab', 'acd'}
    powset = powerset(range(len(attrs)))
    pprems = {B for B in powset if defs.is_proper_premise(B, crosses_per_columns)}
    pprems_verb = {''.join([attrs[m] for m in sorted(B)]) for B in pprems}
    assert pprems_verb == pprems_true


def test_is_minimal_gen():
    crosses_per_columns = [
        {1, 2},  # a
        {3, 4},  # b
        {2, 3, 4},  # c
        {1, 4},  # d
        set(),  # e
    ]
    attrs = 'abcde'
    mingens_true = {'a', 'c', 'd', 'ac', 'ad', '', 'bd', 'b', 'e', 'cd', 'ab', 'acd'}
    powset = powerset(range(len(attrs)))
    mingens = {B for B in powset if defs.is_minimal_gen(B, crosses_per_columns)}
    mingens_verb = {''.join([attrs[m] for m in sorted(B)]) for B in mingens}
    assert mingens_verb == mingens_true


def test_is_minimum_gen():
    crosses_per_columns = [
        {1, 2},  # a
        {3, 4},  # b
        {2, 3, 4},  # c
        {1, 4},  # d
        set(),  # e
    ]
    attrs = 'abcde'
    mingens_true = {'a', 'c', 'd', 'ac', 'ad', '', 'bd', 'b', 'e', 'cd'}
    powset = list(powerset(range(len(attrs))))

    mingens = {B for B in powset if defs.is_minimum_gen(B, crosses_per_columns)}
    mingens_verb = {''.join([attrs[m] for m in sorted(B)]) for B in mingens}
    assert mingens_verb == mingens_true

    minimal_gens = {'a', 'c', 'd', 'ac', 'ad', '', 'bd', 'b', 'e', 'cd', 'ab', 'acd'}
    minimal_gens_unverb = {frozenset({attrs.index(m) for m in D}) for D in minimal_gens}
    mingens = {B for B in powset if defs.is_minimum_gen(B, crosses_per_columns, minimal_gens_unverb)}
    mingens_verb = {''.join([attrs[m] for m in sorted(B)]) for B in mingens}
    assert mingens_verb == mingens_true


def test_is_key():
    crosses_per_columns = [
        {1, 2},  # a
        {3, 4},  # b
        {2, 3, 4},  # c
        {1, 4},  # d
        set(),  # e
    ]
    attrs = 'abcde'
    keys_true = {'a', 'c', 'd', 'ac', 'ad', '', 'bd', 'b', 'e', 'cd', 'ab', 'acd'}
    powset = powerset(range(len(attrs)))
    keys = {B for B in powset if defs.is_key(B, crosses_per_columns)}
    keys_verb = {''.join([attrs[m] for m in sorted(B)]) for B in keys}
    assert keys_verb == keys_true


def test_is_passkey():
    crosses_per_columns = [
        {1, 2},  # a
        {3, 4},  # b
        {2, 3, 4},  # c
        {1, 4},  # d
        set(),  # e
    ]
    attrs = 'abcde'
    pkeys_true = {'a', 'c', 'd', 'ac', 'ad', '', 'bd', 'b', 'e', 'cd'}
    powset = list(powerset(range(len(attrs))))

    pkeys = {B for B in powset if defs.is_passkey(B, crosses_per_columns)}
    pkeys_verb = {''.join([attrs[m] for m in sorted(B)]) for B in pkeys}
    assert pkeys_verb == pkeys_true

    keys = {'a', 'c', 'd', 'ac', 'ad', '', 'bd', 'b', 'e', 'cd', 'ab', 'acd'}
    keys_unverb = {frozenset({attrs.index(m) for m in D}) for D in keys}
    pkeys = {B for B in powset if defs.is_passkey(B, crosses_per_columns, keys_unverb)}
    pkeys_verb = {''.join([attrs[m] for m in sorted(B)]) for B in pkeys}
    assert pkeys_verb == pkeys_true



"""
This module contains functions to test whether
a set of attributes belongs to a given characteristic attribute set class
(e.g. pseudo-intents, minimal generators, etc.)
"""
from typing import FrozenSet, Union
from bitarray import frozenbitarray as fbarray


from .base_functions import closure, is_psubset_of, powerset


def is_closed(B: Union[FrozenSet[int], fbarray], crosses_per_columns: Union[list[FrozenSet[int]], list[fbarray]]) -> bool:
    """Test whether `B` is closed w.r.t. `crosses_per_columns`"""
    return B == set(closure(B, crosses_per_columns))


def is_pseudo_intent(
        P: FrozenSet[int],
        crosses_per_columns: Union[list[FrozenSet[int]], list[fbarray]],
        sub_pseudo_intents: list[FrozenSet[int]] = None,
        is_closed_: bool = None,
) -> bool:
    """Test whether `P` is a pseudo-intent w.r.t. `crosses_per_columns`

    An attribute set P is a pseudo-intent iff
    * P is not closed, and
    * for every smaller pseudo-intent Q, the closure of Q is in P
    """
    if is_closed_ is None:
        is_closed_ = is_closed(P, crosses_per_columns)

    if is_closed_:
        return False

    if sub_pseudo_intents is None:
        sub_pseudo_intents = []
        for D in powerset(P):
            if frozenset(D) == P:
                continue
            if is_pseudo_intent(D, crosses_per_columns, sub_pseudo_intents):
                sub_pseudo_intents.append(D)

    answer = all(
        is_psubset_of(frozenset(closure(Q, crosses_per_columns)), P)
        for Q in sub_pseudo_intents
        if is_psubset_of(Q, P)
    )
    return answer


def is_proper_premise(
        Q: FrozenSet[int],
        crosses_per_columns: list[FrozenSet[int]],
        is_closed_: bool = None
) -> bool:
    """Test whether `Q` is a proper premise w.r.t. `crosses_per_columns`


    An attribute set Q is a proper premise
    * Q is not closed, and
    * Q | closure(Q-{n1}) | ... | closure(Q - {nk}) != closure(Q), where n1, ..., nk are elements of Q
    """
    if is_closed_ is None:
        is_closed_ = is_closed(Q, crosses_per_columns)
    if is_closed_:
        return False

    intent = set(closure(Q, crosses_per_columns))
    union_closure = set(Q)
    for n in Q:
        union_closure |= set(closure(Q-{n}, crosses_per_columns))
    return union_closure != intent


def is_minimal_gen(D: FrozenSet[int], crosses_per_columns: list[FrozenSet[int]]) -> bool:
    """Test whether `D` is a minimal generator w.r.t. `crosses_per_columns`

    An attribute set D is a minimal generator if there is no subset of D with the same closure as D
    """
    intent = set(closure(D, crosses_per_columns))
    return all(set(closure(D-{m}, crosses_per_columns)) != intent for m in D)


def is_minimum_gen(
        D: FrozenSet[int],
        crosses_per_columns: list[FrozenSet[int]],
        minimal_gens: list[FrozenSet[int]] = None,
        is_minimal_gen_: bool = None
) -> bool:
    """Test whether `D` is a minimum generator w.r.t. `crosses_per_columns`


    An attribute set D is a minimum generator if there is no attribute set of smaller size with the same closure as D
    """
    if is_minimal_gen_ is None:
        is_minimal_gen_ = is_minimal_gen(D, crosses_per_columns)
    if is_minimal_gen_ is False:
        return False

    if minimal_gens is None:
        descr_iterator = powerset(range(len(crosses_per_columns)))
    else:
        descr_iterator = sorted(minimal_gens, key=lambda E: len(E))
    intent = set(closure(D, crosses_per_columns))
    first_mingen = next(E for E in descr_iterator if set(closure(E, crosses_per_columns)) == intent)
    return len(D) == len(first_mingen)


def is_key(D: frozenset, crosses_per_columns: list[frozenset], is_minimal_gen_=None) -> bool:
    """Test whether `D` is a key w.r.t. `crosses_per_columns` (the same as minimal generator)"""
    return is_minimal_gen(D, crosses_per_columns) if is_minimal_gen_ is None else is_minimal_gen_


def is_passkey(
        D: FrozenSet[int],
        crosses_per_columns: list[FrozenSet[int]],
        keys: list[FrozenSet[int]] = None,
        is_minimum_gen_=None
) -> bool:
    """Test whether `D` is a passkey w.r.t. `crosses_per_columns` (the same as minimum generator)"""
    return is_minimum_gen(D, crosses_per_columns, keys) if is_minimum_gen_ is None else is_minimum_gen_



from bitarray import frozenbitarray as fbarray

from caspailleur import indices as idxs


def test_delta_stability_by_description():
    attr_extents = [
        fbarray('100101'),
        fbarray('011010'),
        fbarray('100101'),
        fbarray('010010'),
        fbarray('001000'),
        fbarray('111101'),
        fbarray('000010')
    ]

    delta_stab = idxs.delta_stability_by_description(set(), attr_extents)
    assert delta_stab == 1

    delta_stab = idxs.delta_stability_by_description(fbarray('0'*7), attr_extents)
    assert delta_stab == 1

    delta_stab = idxs.delta_stability_by_description([5], attr_extents)
    assert delta_stab == 2
    delta_stab = idxs.delta_stability_by_description(fbarray('0000010'), attr_extents)
    assert delta_stab == 2



"""Module with high-level functions to work with Caspailleur

The proposed functions are:
* iter_descriptions(data) to iterate all possible descriptions (and their characteristics) in the data
* mine_descriptions(data) to get the Pandas DataFrame with all possible descriptions in the data
    (note that "the number of all descriptions" = 2**"the number of binary attributes",
    so the resulting DataFrame might be huge even for small data)
* mine_concepts(data) to get all the concepts (and their characteristics) in the data
* mine_implications(data, basis_name) to get a basis of implications for the data
"""
from typing import Iterator, Iterable, Literal, Union, Optional, get_args
import pandas as pd
from bitarray import frozenbitarray as fbarray

from .base_functions import powerset, extension, intention
from .io import ContextType, to_named_bitarrays, transpose_context, isets2bas, verbalise, to_absolute_number
from .order import topological_sorting, sort_intents_inclusion, inverse_order
from . import indices as idxs
from . import definitions
from . import mine_equivalence_classes as mec, implication_bases as ibases


MINE_DESCRIPTION_COLUMN = Literal[
    "description", "extent", "intent",
    "support", "delta_stability",
    "is_closed", "is_key", "is_passkey", "is_proper_premise", "is_pseudo_intent"
]

MINE_CONCEPTS_COLUMN = Literal[
    "extent", "intent", "support", "delta_stability",
    "keys", "passkeys", "proper_premises", "pseudo_intents",
    "previous_concepts", "next_concepts", "sub_concepts",  "super_concepts",
]

MINE_IMPLICATIONS_COLUMN = Literal[
    "premise", "conclusion", "conclusion_full", "extent", "support"
]

BASIS_NAME = Literal[
    "Proper Premise", "Canonical Direct", "Karell",
    "Pseudo-Intent", "Canonical", "Duquenne-Guigues", "Minimum",
]


def _setup_colnames_to_compute(
        all_columns,
        columns_to_compute: Union[list[str], Literal['all'], None],
        dependencies: dict[Union[str, tuple[str, bool]], set[str]], return_all_computed: bool
) -> tuple[set[str], list[str]]:
    columns_to_return = list(get_args(all_columns))
    if columns_to_compute is not None and columns_to_compute != 'all':
        columns_to_return = list(columns_to_compute)
    columns_to_compute = set(columns_to_return)

    assert columns_to_compute <= set(get_args(all_columns)), \
        f"The following elements were asked for but cannot be computed {columns_to_compute - set(all_columns)}. " \
        f"However, only the following columns can be chosen: {all_columns}."

    for premise, conclusion in dependencies.items():
        if not isinstance(premise, str):
            if not premise[1]:
                continue
            premise = premise[0]
        columns_to_compute.update(conclusion if premise in columns_to_compute or premise == '' else set())

    if return_all_computed:
        columns_to_return += sorted(set(columns_to_compute) - set(columns_to_return), key=list(get_args(all_columns)).index)
    return columns_to_compute, columns_to_return


def iter_descriptions(
        data: ContextType,
        to_compute: Union[list[MINE_DESCRIPTION_COLUMN], Literal['all']] = 'all'
) -> Iterator[dict]:
    """Iterate all possible descriptions in the data one by one

    Parameters
    ----------
    data: ContextType
        Data in a format, supported by Caspailleur.io module.
        For example, Pandas DataFrame with binary values,
        or list of lists of strings (where every list of strings represents an itemset).
    to_compute: list[MINE_DESCRIPTION_COLUMN] or 'all'
        A list of characteristics to compute (by default: "all")
        The list of all possible characteristics is defined in caspailleur.MINE_DESCRIPTION_COLUMN value.

    Returns
    -------
    descriptions: Iterator[dict]
        Iterator of descriptions. Every description and its characteristics are represented with a dictionary
        where the keys are defined by `to_compute` parameter.


    Notes
    -----
    Every description is a subset of attributes,
    and the maximal amount of possible descriptions is 2**(number of attributes).
    Thus, the resulting iterator will contain the exponential amount of elements,
    unless you define some additional early stopping outside of this function.
    """
    bitarrays, objects, attributes = to_named_bitarrays(data)
    attr_extents = transpose_context(bitarrays)
    to_compute, cols_to_return = _setup_colnames_to_compute(MINE_DESCRIPTION_COLUMN, to_compute, {}, True)

    sub_pseudo_intents = []

    def get_vals_for_column(column_name: MINE_DESCRIPTION_COLUMN):
        if column_name == 'description':
            return set(verbalise(description_idxs, attributes))
        if column_name == 'extent':
            return set(verbalise(extent_ba, objects))
        if column_name == 'intent':
            return set(verbalise(definitions.closure(description_idxs, attr_extents), attributes))
        if column_name == 'support':
            return idxs.support_by_description(description_idxs, attr_extents, extent_ba)
        if column_name == 'delta_stability':
            return idxs.delta_stability_by_description(description_idxs, attr_extents)
        if column_name == 'is_closed':
            return definitions.is_closed(description_idxs, attr_extents)
        if column_name == 'is_key':
            return definitions.is_key(description_idxs, attr_extents)
        if column_name == 'is_passkey':
            return definitions.is_passkey(description_idxs, attr_extents)
        if column_name == 'is_proper_premise':
            return definitions.is_proper_premise(description_idxs, attr_extents)
        if column_name == 'is_pseudo_intent':
            return definitions.is_pseudo_intent(description_idxs, attr_extents, sub_pseudo_intents)
        raise NotImplementedError(f"Something's gone wrong in the code, asked for column: {column_name}")

    for description_idxs in powerset(range(len(attributes))):
        extent_ba = fbarray(extension(description_idxs, attr_extents))
        stats = {colname: get_vals_for_column(colname) for colname in cols_to_return}

        yield stats
        if 'is_pseudo_intent' in cols_to_return and stats['is_pseudo_intent']:
            sub_pseudo_intents.append(description_idxs)


def mine_descriptions(
        data: ContextType,
        min_support: Union[int, float] = 0,
        to_compute: Union[list[MINE_DESCRIPTION_COLUMN], Literal['all']] = 'all',
        return_every_computed_column: bool = False
) -> pd.DataFrame:
    """Mine all frequent descriptions and their characteristics

    The frequency of a description is defined by its "support" index,
    which is the amount (or the percentage) of objects described by the description.

    Parameters
    ----------
    data: ContextType
        Data in a format, supported by Caspailleur.io module.
        For example, Pandas DataFrame with binary values,
        or list of lists of strings (where every list of strings represents an itemset).
    min_support: int or float
        Minimal value of the support for a description, i.e. how many objects it describes.
        Can be defined by an integer (so the number of objects) or by a float (so the percentage of objects).
    to_compute: list[MINE_DESCRIPTION_COLUMN] or 'all'
        A list of characteristics to compute (by default: "all")
        The list of all possible characteristics is defined in caspailleur.MINE_DESCRIPTION_COLUMN value.
    return_every_computed_column: bool
        A flag whether to return every computed column or only the ones defined by `to_compute` parameter.
        For example, the computation of column 'is_proper_premise' requires the computation of column 'is_key'.
        Thus, if you have asked for column 'is_proper_premise' you can get the column 'is_key' ``for free''.

    Returns
    -------
    descriptions_df: pandas.DataFrame
        Pandas DataFrame where every row represents a description, and every column represents its characteristics
        defined by `to_compute` and `return_every_computed_column` parameters.

    Notes
    -----
    If you want to look at only the most stable descriptions, then use "mine_concepts" function,
    as all stable descriptions are concept intents.

    Every description is a subset of attributes.
    So the maximal amount of possible descriptions is 2**(number of attributes).
    Make sure to specify the `min_support` parameter when dealing with the data with dozens of attributes,
    otherwise you might end up with a resulting DataFrame of hundreds of thousands rows.
    """
    ####################################################
    # Computing what columns and parameters to compute #
    ####################################################
    col_dependencies: dict[MINE_DESCRIPTION_COLUMN, set[MINE_DESCRIPTION_COLUMN]] = {
        'is_pseudo_intent': {'is_proper_premise', 'intent'},
        'is_proper_premise': {'intent', 'is_key'},
        'is_key': {'intent'},
        'is_passkey': {'intent'},
        'is_closed': {'intent'},
        ('', min_support > 0): {'support'},  # always compute support if min_support > 0
        'support': {'extent'},
        'delta_stability': {'extent'},
        'intent': {'extent'},
    }
    to_compute, cols_to_return = _setup_colnames_to_compute(
        MINE_DESCRIPTION_COLUMN, to_compute, col_dependencies, return_every_computed_column)

    ################################
    # Compute the required columns #
    ################################
    bitarrays, objects, attributes = to_named_bitarrays(data)
    attr_extents = transpose_context(bitarrays)
    n_objects = len(objects)
    min_support = to_absolute_number(min_support, n_objects)

    descriptions_ba = list(isets2bas(powerset(range(len(attributes))), len(attributes)))
    if 'extent' in to_compute:
        extents_ba = [fbarray(extension(descr, attr_extents)) for descr in descriptions_ba]
    if min_support > 0:
        descriptions_ba, extents_ba = zip(*[(descr, ext) for descr, ext in zip(descriptions_ba, extents_ba)
                                              if ext.count() >= min_support])
    column_description = [verbalise(descr, attributes) for descr in descriptions_ba]
    if 'extent' in to_compute:
        column_extent = [verbalise(extent, objects) for extent in extents_ba]
    if 'intent' in to_compute:
        intents_ba = mec.list_intents_via_LCM(bitarrays, min_supp=min_support)
        ext_int_map = {fbarray(extension(intent, attr_extents)): intent_i for intent_i, intent in enumerate(intents_ba)}
        column_intent = [verbalise(intents_ba[ext_int_map[extent]], attributes) for extent in extents_ba]
    if 'is_key' in to_compute:
        keys_ba = mec.list_keys(intents_ba)
    if 'is_passkey' in to_compute:
        passkeys_ba = mec.list_passkeys(intents_ba)
    if 'is_proper_premise' in to_compute:
        ppremises_ba = dict(ibases.iter_proper_premises_via_keys(intents_ba, keys_ba))
    if 'is_pseudo_intent' in to_compute:
        pintents_ba = dict(ibases.list_pseudo_intents_via_keys(ppremises_ba.items(), intents_ba))
    if 'support' in to_compute:
        column_support = [extent.count() for extent in extents_ba]
    if 'delta_stability' in to_compute:
        column_delta_stability = [idxs.delta_stability_by_description(descr, attr_extents, extent)
                                  for descr, extent in zip(descriptions_ba, extents_ba)]
    if 'is_closed' in to_compute:
        column_is_closed = [intents_ba[ext_int_map[extent]] == descr_ba
                            for descr_ba, extent in zip(descriptions_ba, extents_ba)]
    if 'is_key' in to_compute:
        column_is_key = [descr_ba in keys_ba for descr_ba in descriptions_ba]
    if 'is_passkey' in to_compute:
        column_is_passkey = [descr_ba in passkeys_ba for descr_ba in descriptions_ba]
    if 'is_proper_premise' in to_compute:
        column_is_proper_premise = [descr_ba in ppremises_ba for descr_ba in descriptions_ba]
    if 'is_pseudo_intent' in to_compute:
        column_is_pseudo_intent = [descr_ba in pintents_ba for descr_ba in descriptions_ba]

    locals_ = locals()
    return pd.DataFrame({f: locals_[f"column_{f}"] for f in cols_to_return}).rename_axis('description_id')


def mine_concepts(
        data: ContextType,
        to_compute: Union[list[MINE_CONCEPTS_COLUMN], Literal['all']] = 'all',
        return_every_computed_column: bool = False,
        min_support: Union[int, float] = 0,
        min_delta_stability: Union[int, float] = 0, n_stable_concepts: Optional[int] = None,
        use_tqdm: bool = False,
) -> pd.DataFrame:
    """Compute the frequent concepts in the data

    Parameters
    ----------
    data: ContextType
        Data in a format, supported by Caspailleur.io module.
        For example, Pandas DataFrame with binary values,
        or list of lists of integers (where every list of integers represents an itemset).
    to_compute: list[MINE_CONCEPT_COLUMN] or 'all'
        A list of characteristics to compute (by default, compute 'all' possible columns)
        The list of all possible characteristics is defined in caspailleur.MINE_CONCEPT_COLUMN value.
    return_every_computed_column: bool
        A flag whether to return every computed column or only the ones defined by `to_compute` parameter.
        For example, the computation of column 'is_proper_premise' requires the computation of column 'is_key'.
        Thus, if you have asked for column 'is_proper_premise' you can get the column 'is_key' ``for free''.
    min_support: int or float
        Minimal value of the support for a concept, i.e. how many objects it describes.
        Can be defined by an integer (so the number of objects) or by a float (so the percentage of objects).
    min_delta_stability: int or float
        Minimal value of the delta-stability of a concept,
        i.e. the minimal amount of objects a concept will lose when made a bit more precise.
        Can be defined by an integer (so the number of objects) or by a float (so the percentage of objects).
    n_stable_concepts: int or None
        Select only `n` concepts with the highest (presumably) delta-stability.
        The parameter can be used together with or instead of `min_delta_stability`
        when the exact value the of required min_delta_stability is not known.
        The found `n` stable concepts are not necessarily the concepts with the _highest_ delta stability.
        Yet their delta-stability is high enough.
    use_tqdm: bool
        A flag whether to use tqdm progress bar for long computations.
        Is used for computing pseudo-intents.

    Returns
    -------
    descriptions_df: pandas.DataFrame
        Pandas DataFrame where every row represents a concept, and every column represents its characteristics
        defined by `to_compute` and `return_every_computed_column` parameters.

    Notes
    -----
    The number of concepts might be exponential to the number of objects and attributes in the data.
    Make sure to specify the `min_support` parameter when dealing with the data with dozens of attributes,
    otherwise you might end up with a resulting DataFrame of hundreds of thousands rows.

    """
    ##################################################
    # Computing what columns and parameters to compute
    ##################################################
    # whether to compute all concepts whose support is higher min_support
    compute_only_stable_concepts = min_delta_stability != 0 or n_stable_concepts is not None
    col_dependencies: dict[MINE_CONCEPTS_COLUMN, set[MINE_CONCEPTS_COLUMN]] = {
        'pseudo_intents': {'proper_premises', 'intent'},
        'proper_premises': {'intent', 'keys'},
        'keys': {'intent'},
        'passkeys': {'intent'},
        ('keys', compute_only_stable_concepts): {'extent'},  # extents are not required when computing join semilattice
        ('passkeys', compute_only_stable_concepts): {'extent'},
        'support': {'extent'},
        'delta_stability': {'intent', 'extent'},
        'extent': {'intent'},
        'super_concepts': {'sub_concepts'},
        'sub_concepts': {'previous_concepts'},
        'next_concepts': {'previous_concepts'},
        'previous_concepts': {'intent'},
    }
    to_compute, cols_to_return = _setup_colnames_to_compute(
        MINE_CONCEPTS_COLUMN, to_compute, col_dependencies, return_every_computed_column)

    #################################
    # Running the (long) computations
    #################################
    bitarrays, objects, attributes = to_named_bitarrays(data)
    attr_extents = transpose_context(bitarrays)
    bitarrays, attr_extents = list(map(fbarray, bitarrays)), list(map(fbarray, attr_extents))

    def verbalise_descriptions(bas):
        return [verbalise(ba, attributes) for ba in bas]

    def group_by_concept(pairs: Iterable[tuple[fbarray, int]], n_cncpts: int) -> list[list[set[str]]]:
        per_concept = [[] for _ in range(n_cncpts)]
        for ba, cncpt_i in pairs:
            per_concept[cncpt_i].append(ba)
        return per_concept

    if 'intent' in to_compute:
        if compute_only_stable_concepts:
            n_objects = len(objects)
            stable_extents = mec.list_stable_extents_via_gsofia(
                attr_extents, n_objects, min_delta_stability, n_stable_concepts,
                min_supp=to_absolute_number(min_support, n_objects), n_attributes=len(attr_extents)
            )
            intents_ba = [intention(extent, attr_extents) for extent in stable_extents]
        else:
            intents_ba = mec.list_intents_via_LCM(bitarrays, min_supp=min_support)

        intents_ba = topological_sorting(intents_ba)[0]
        n_concepts = len(intents_ba)
        column_intent = verbalise_descriptions(intents_ba)
    if 'extent' in to_compute:
        extents_ba = [fbarray(extension(intent, attr_extents)) for intent in intents_ba]
        column_extent = [verbalise(extent, objects) for extent in extents_ba]
    if 'keys' in to_compute:
        if compute_only_stable_concepts:
            keys_ba = mec.list_keys_for_extents(extents_ba, attr_extents)
        else:
            keys_ba = mec.list_keys(intents_ba)

        column_keys = [verbalise_descriptions(keys) for keys in group_by_concept(keys_ba.items(), n_concepts)]
    if 'passkeys' in to_compute:
        if compute_only_stable_concepts:
            passkeys_ba = mec.list_passkeys_for_extents(extents_ba, attr_extents)
        else:
            passkeys_ba = mec.list_passkeys(intents_ba)
        column_passkeys = [verbalise_descriptions(pkeys) for pkeys in group_by_concept(passkeys_ba.items(), n_concepts)]
    if 'proper_premises' in to_compute:
        ppremises_ba = dict(ibases.iter_proper_premises_via_keys(intents_ba, keys_ba))
        column_proper_premises = [verbalise_descriptions(pps) for pps in group_by_concept(ppremises_ba.items(), n_concepts)]
    if 'pseudo_intents' in to_compute:
        pintents_ba = dict(ibases.list_pseudo_intents_via_keys(
            ppremises_ba.items(), intents_ba, use_tqdm=use_tqdm, n_keys=len(ppremises_ba)))
        column_pseudo_intents = [verbalise_descriptions(pis) for pis in group_by_concept(pintents_ba.items(), n_concepts)]
    if 'support' in to_compute:
        column_support = [extent.count() for extent in extents_ba]
    if 'delta_stability' in to_compute:
        column_delta_stability = [idxs.delta_stability_by_description(descr, attr_extents, extent_ba)
                                  for descr, extent_ba in zip(intents_ba, extents_ba)]

    # Columns for order on concepts
    if 'previous_concepts' in to_compute:
        previous_concepts, sub_concepts = sort_intents_inclusion(intents_ba, return_transitive_order=True)
    if 'next_concepts' in to_compute:
        next_concepts = inverse_order(previous_concepts)
    if 'super_concepts' in to_compute:
        super_concepts = inverse_order(sub_concepts)
    for colname in ['super_concepts', 'sub_concepts', 'next_concepts', 'previous_concepts']:
        if colname in cols_to_return:
            locals()[f"column_{colname}"] = [set(ba.search(True)) for ba in locals()[f"{colname}"]]

    locals_ = locals()
    return pd.DataFrame({f: locals_[f"column_{f}"] for f in cols_to_return}).rename_axis('concept_id')


def mine_implications(
        data: ContextType, basis_name: BASIS_NAME = 'Proper Premise',
        unit_base: bool = False,
        to_compute: Optional[Union[list[MINE_IMPLICATIONS_COLUMN], Literal['all']]] = 'all',
        return_every_computed_column: bool = False,
        min_support: Union[int, float] = 0,
        min_delta_stability: Union[int, float] = 0 , n_stable_concepts: Optional[int] = None
) -> pd.DataFrame:
    """Compute an implication basis (i.e. a set of non-redundant implications) for the given data

    An implication is a pair A -> B meaning that
     whenever all attributes from A are found in a description, this description also contains every attribute from B.
    Here, set of attributes A is called a premise, set of attributes B is called a conclusion.

    Parameters
    ----------
    data: ContextType
        Data in a format, supported by Caspailleur.io module.
        For example, Pandas DataFrame with binary values,
        or list of lists of strings (where every list of strings represents an itemset).
    basis_name: BASIS_NAME
        The name of the basis two compute. The possible values are provided in caspailleur.BASIS_NAME variable
         and discussed in the Notes section below.
    unit_base: bool
        A flag whether to return the "unit" version of the basis,
         where every implication's conclusion consists of a single attribute.
        See the Notes section below for more examples.
    to_compute: list[MINE_IMPLICATIONS_COLUMN] or 'all'
        A list of characteristics to compute (by default, compute 'all')
        The list of all possible characteristics is defined in caspailleur.MINE_IMPLICATION_COLUMN value.
    return_every_computed_column: bool
        A flag whether to return every computed column or only the ones defined by `to_compute` parameter.
        For example, the computation of column 'support' requires the computation of column 'extent'.
        Thus, if you have asked for column 'support' you can get the column 'extent' ``for free''.
    min_support: int or float
        Minimal value of the support for an implication, i.e. how many objects it describes.
        Can be defined by an integer (so the number of objects) or by a float (so the percentage of objects).
    min_delta_stability: int or float
        Minimal delta stability of concepts, whose minimal/maximal descriptions serve as premises/conclusions
         of implications. Raising up the value of min_delta_stability might be useful for finding implications in
          big data, where finding all concepts is not feasible.
        The value can be provided as an integer (measuring delta-stability with the number of objects) or as a float
         (measuring delta-stability with the percentage of objects w.r.t. total number of objects in the data).
        NOTE: The use of min_delta_stability for finding implication was not extensively studied in the literature.
        The resulting set of implication should be valid, but might not reflect some important implications.
    n_stable_concepts: int or None
        Parameter analogous to min_delta_stability,  except that it sets up the maximal value of concepts
         with high delta stability, and not the threshold for minimal delta stability.
        Can be either an integer (representing maximal number of high-stability concepts) or None (for all concepts).

    Returns
    -------
    descriptions_df: pandas.DataFrame
        Pandas DataFrame where every row represents an implication, and every column represents its characteristics
        defined by `to_compute` parameter.

    Notes
    -----
    *Basis name*
    There are two implication bases that can be mined via Caspailleur.
    One is called "Canonical Direct" basis and sometimes referred to as "Proper Premise" basis or "Karell" basis.
    The other is called "Canonical" basis and sometimes referred to as "Duquenne-Guigues",
     "Minimum" or "Pseudo-Intent" basis.

    Canonical Direct basis is "direct",
    i.e. one can find the closure of a subset of attributes with one pass over the implications.
    Also, every premise of the Canonical Direct basis is a minimal description of all attributes
    in the conclusion of the premise. Thus, no subset of a premise implies the same conclusion.

    Canonical basis contains the smallest amount of implication possible.
    Thus, it can be considered as the most dense and the most concise way to represent the data.
    However, such basis is not "direct"
    and some premises of this basis are not minimal subsets of attributes corresponding to their conclusions.

    Note that the Caspailleur's algorithm for computing the bases are not proven to be State-of-the-Art,
    as they were never extensively tested versus the other algorithms.
    But, they work fast enough for daily occasions.


    *Unit base*
    Unit and non-unit bases are two ways of representing the same set of implications.
    In the unit base, every implication's conclusion consist of a single attribute,
     but there can be many implications with the same premise.
    In the non-unit base, every implication's premise is unique, but there can be many attributes in the conclusion.

    For example, non unit-base:
    {a} -> {b, c}
    {d, e} -> {f}

    and the equivalent unit base:
    {a} -> b
    {a} -> c
    {d, e} -> f
    """
    assert basis_name in get_args(BASIS_NAME), \
        f"You asked for '{basis_name}' basis. But only the following bases are supported: {BASIS_NAME}"
    if basis_name in {'Canonical Direct', "Karell"}:
        basis_name = 'Proper Premise'
    if basis_name in {'Canonical', 'Duquenne-Guigues', 'Minimum'}:
        basis_name = 'Pseudo-Intent'

    dependencies = {
        "support": {'extent'},
    }

    to_compute, cols_to_return = _setup_colnames_to_compute(
        MINE_IMPLICATIONS_COLUMN, to_compute, dependencies, return_all_computed=return_every_computed_column)

    ############################
    # Compute the (unit) basis #
    ############################
    bitarrays, objects, attributes = to_named_bitarrays(data)
    attr_extents = transpose_context(bitarrays)
    bitarrays, attr_extents = list(map(fbarray, bitarrays)), list(map(fbarray, attr_extents))

    n_objects = len(objects)
    min_support, min_delta_stability = [to_absolute_number(v, n_objects) for v in [min_support, min_delta_stability]]
    compute_only_stable_concepts = min_delta_stability != 0 or n_stable_concepts is not None

    extents_ba, int_ext_map = None, None
    if compute_only_stable_concepts:
        intents_ba = mec.list_intents_via_LCM(bitarrays, min_supp=min_support)
        keys_ba = mec.list_keys(intents_ba)
    else:  # min_delta_stability > 0, so searching for only some stable concepts
        stable_extents = list(mec.list_stable_extents_via_gsofia(
            attr_extents, n_objects, min_delta_stability,
            n_stable_extents=n_stable_concepts, min_supp=min_support, n_attributes=len(attr_extents)))
        intents_ba = [intention(extent, attr_extents) for extent in stable_extents]
        int_ext_map = dict(zip(intents_ba, stable_extents))
        intents_ba = topological_sorting(intents_ba)[0]
        extents_ba = [int_ext_map[intent] for intent in intents_ba]
        keys_ba = mec.list_keys_for_extents(extents_ba, attr_extents)

    ppremises_ba = ibases.iter_proper_premises_via_keys(
        intents_ba, keys_ba,
        all_frequent_keys_provided=not compute_only_stable_concepts)
    ppremises_ba = sorted(ppremises_ba, key=lambda prem_intent: (prem_intent[1], prem_intent[0].count()))
    basis = ppremises_ba
    if basis_name == 'Pseudo-Intent':
        basis = ibases.list_pseudo_intents_via_keys(ppremises_ba, intents_ba)

    # compute pseudo-closures to reduce the conclusions by attributes implied by other implications
    if 'conclusion' in to_compute:
        pseudo_closures = [
            ibases.saturate(premise, [impl for impl in basis[:impl_i] if impl[1] < intent_i], intents_ba)
            for impl_i, (premise, intent_i) in enumerate(basis)
        ]
        basis = [(premise, intents_ba[intent_i] & ~pintent, intent_i)
                 for (premise, intent_i), pintent in zip(basis, pseudo_closures)]
    else:
        basis = [(premise, None, intent_i) for premise, intent_i in basis]

    if unit_base and 'conclusion' in to_compute:
        single_attrs = to_named_bitarrays([{i} for i in range(len(attributes))])[0]
        basis = [(premise, single_attrs[attr_i], intent_i)
                 for premise, conclusion, intent_i in basis for attr_i in conclusion.search(True)]
    premises, conclusions, intents_idxs = zip(*basis)

    ####################################
    # Prepare values for the dataframe #
    ####################################
    def verbalise_descriptions(bas):
        return [verbalise(ba, attributes) for ba in bas]

    if 'premise' in to_compute:
        column_premise = verbalise_descriptions(premises)
    if 'conclusion' in to_compute:
        column_conclusion = verbalise_descriptions(conclusions)
        if unit_base:
            column_conclusion = [list(conclusion)[0] for conclusion in column_conclusion]
    if 'conclusion_full' in to_compute:
        column_conclusion_full = verbalise_descriptions(map(intents_ba.__getitem__, intents_idxs))
    if "extent" in to_compute:
        if compute_only_stable_concepts:  # otherwise, extents are already computed
            extents_ba = [fbarray(extension(intent, attr_extents)) for intent in intents_ba]
            int_ext_map = dict(zip(intents_ba, extents_ba))
    if "extent" in cols_to_return:
        column_extent = [verbalise(int_ext_map[intents_ba[intent_i]], objects) for intent_i in intents_idxs]
    if "support" in cols_to_return:
        column_support = [int_ext_map[intents_ba[intent_i]].count() for intent_i in intents_idxs]

    locals_ = locals()
    return pd.DataFrame({f: locals_[f"column_{f}"] for f in cols_to_return}).rename_axis('implication_id')



from functools import reduce
from itertools import chain, combinations
from typing import Iterable, Iterator, Union, Any

import deprecation
from bitarray import frozenbitarray as fbarray, bitarray

from . import io


################
# Math notions #
################
def powerset(iterable: Iterable[Any]) -> Iterable[frozenset[Any]]:
    """Iterate through all subsets of elements of set `iterable`

    Examples
    --------
    powerset({1,2,3}) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    """
    s = list(iterable)
    return map(frozenset, chain.from_iterable(combinations(s, r) for r in range(len(s)+1)))


def is_subset_of(A: Union[set[int], fbarray], B: Union[set[int], fbarray]) -> bool:
    """Test whether `A` is a subset of `B`"""
    return A & B == A


def is_psubset_of(A: Union[set[int], fbarray], B: Union[set[int], fbarray]) -> bool:
    """Test whether `A` is a proper subset of `B`"""
    return (A & B == A) and A != B


def maximal_extent(crosses_per_columns: Union[list[set], list[bitarray]]) -> Union[set, bitarray]:
    """Return the whole set of objects from `crosses_per_columns` data representation"""
    first_column = crosses_per_columns[0]
    if isinstance(first_column, bitarray):
        return first_column | (~first_column)
    if any(column and isinstance(next(iter(column)), int) for column in crosses_per_columns):
        n_rows = max(max(column) for column in crosses_per_columns if column) + 1
        return type(first_column)(range(n_rows))

    all_attrs = reduce(first_column.__or__ , crosses_per_columns)
    return type(first_column)(all_attrs)


def extension(description: Union[Iterable[int], bitarray], crosses_per_columns: Union[list[set], list[bitarray]])\
        -> Union[set[int], bitarray]:
    """Select the indices of rows described by `description`"""
    column_type = type(crosses_per_columns[0])
    description = description.itersearch(True) if isinstance(description, bitarray) else description

    total_extent = column_type(maximal_extent(crosses_per_columns))
    extent = reduce(column_type.__and__, (crosses_per_columns[attr_i] for attr_i in description), total_extent)
    return extent


def intention(
        objects: Union[set[int], bitarray], crosses_per_columns: Union[list[set[int]], list[bitarray]]
) -> Union[Iterator[int], bitarray]:
    """Iterate the indices of columns that describe the `objects`"""
    if isinstance(objects, bitarray):
        return type(objects)([is_subset_of(objects, col) for col in crosses_per_columns])

    if isinstance(crosses_per_columns[0], bitarray) and not isinstance(objects, bitarray):
        objects = next(io.isets2bas([objects], len(crosses_per_columns[0])))

    return (m for m, col in enumerate(crosses_per_columns) if is_subset_of(objects, col))


def closure(
        description: Union[Iterable[int], bitarray], crosses_per_columns: Union[list[set[int]], list[bitarray]]
) -> Union[Iterator[int], bitarray]:
    """Iterate indices of all columns who describe the same rows as `description`

    Parameters
    ----------
    description:
        Indices of some columns from `crosses_per_columns`
    crosses_per_columns:
        List of indices of 'True' rows for each column

    Returns
    -------
    intent:
        Indices of All columns with the same intersection as `D`

    """
    result = intention(extension(description, crosses_per_columns), crosses_per_columns)
    if not isinstance(description, bitarray) and isinstance(result, bitarray):
        return (i for i in result.itersearch(True))
    return result


##############################
# Reverse compatibility part #
##############################
@deprecation.deprecated(deprecated_in="0.2.0", removed_in="0.2.1",
                        details="The function is moved to `caspailleur.io` module")
def isets2bas(itemsets: Iterable[Iterable[int]], length: int) -> Iterator[fbarray]:
    """Convert the list of lists of indices of 'True' elements to bitarrays of given length

    Examples
    --------
    isets2bas([ [0, 1], [1,3,4] ], 5) --> [bitarray('01000'), bitarray('01011')]
    """
    return io.isets2bas(itemsets, )
